/*------------------------------------------------------------------------*/

#define COMPSAT_COPYRIGHT \
"(C)2002 Armin Biere, Computer Systems Institute, ETH Zurich, Switzerland"

#define COMPSAT_ID \
"$Id: compsat.c,v 1.8 2003/04/03 19:40:01 biere Exp $"

/*------------------------------------------------------------------------*/

#include "config.h"
#include "compsat.h"
#include "dimacs.h"

/*------------------------------------------------------------------------*/
#ifndef COMPSAT_CHECK
#define NDEBUG
#endif
#include <assert.h>
/*------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <ctype.h>

/*------------------------------------------------------------------------*/

#define COMPSAT_PREFIX		"c compsat "
#define LD_BITS_PER_WORD	5
#define BITS_PER_WORD		(1 << LD_BITS_PER_WORD)
#define MAX_WATCHED		2
#define ISORT_LIMIT		10
#define FALSE			-INT_MAX
#define TRUE			INT_MAX
#define MAX_VAR			(INT_MAX/2)
#define AGING_DIVISOR		4

/*------------------------------------------------------------------------*/

typedef struct Queue Queue;
typedef struct Counter Counter;
typedef struct Counters Counters;
typedef struct Limits Limits;
typedef struct Statistics Statistics;

/*------------------------------------------------------------------------*/
/* The attributes of variables can be allocated in one chunk of memory each
 * or split apart and stored in stacks.  The latter may provide better memory
 * locality and thus faster execution speed on some machines.
 */
#ifndef COMPSAT_SLICE_VARIABLE
typedef struct Variable Variable;
#endif

/*------------------------------------------------------------------------*/
/* In essence we have two types of stacks.  One optimized for space, the
 * other optimized for time.  The assumption for fast stacks is, that a
 * small set of instances is used rather frequently.  Compact stacks are
 * optimized for space and try to minimize the memory usage.  The individual
 * instance of a compact stack is assumed to be used less frequently than
 * the fast stacks but usually occurs in high numbers.  We also have more
 * operations on fast stacks.  Compact stacks are implemented differently
 * for 32 bit machines and 64 bit machines.  The generic implementation
 * works for both, 32 bit and 64 bit machines, but is one pointer sized word
 * larger.  It is also probably slightly faster.  There is a size limit of
 * 1GB for compact stacks in the specific 32 bit implementation, which can
 * be removed by switching to the generic implementation. 
 * 
 * The selection of which stacks is used is done with the command line
 * options of 'configure'.  The necessary code is only included if the
 * flags 'COMPSAT_INCLUDE_{GenericCompactStack,CompactStack32}' are defined.
 */
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_INCLUDE_CompactStack32
/*------------------------------------------------------------------------*/

typedef struct CompactStack32 CompactStack32;

/*------------------------------------------------------------------------*/

struct CompactStack32
{
  unsigned word[2];		/* assume 'unsigned' is 32 bit */
};

/*------------------------------------------------------------------------*/
#ifdef THIS_IS_JUST_A_COMMENT
/*
 * This just explains the bit field layout of 'CompactStack32'.
 */
struct BitFieldVersion_of_CompactStack32
{
  /* Assume 4 byte aligned pointers fit into a 32 bit unsigned word.  For
   * fast retrieval of the 'base' pointer the 30 valid bits of the pointer
   * are assumed to be left aligned.  Then the pointer can be obtained by
   * just masking out the lower 2 bits occupied by 'ldsize_low'.
   */
  unsigned base:30;

  /* The logarithm 'ldsize' of the size of a stack is split into its uppper
   * part 'ldhigh' and into its lower part 'ldlow' such that the whole
   * structure only needs 2 words.
   */
  unsigned ldsize_low:2;

  /*
   * Here starts the second word.
   */

  unsigned ldsize_high:3;

  unsigned large_entries:1;	/* 1 iff contents is 32 bit words */

  /* This restricts the maximal number of bytes of the allocated block to 1
   * GB.  For fast access to 'count' we assume that its 28 bits are right
   * aligned and can just be obtained by masking out the upper 4 bits.
   */
  unsigned count:28;
};
#endif
/*------------------------------------------------------------------------*/
#endif /* ifdef COMPSAT_INCLUDE_CompactStack32 */
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_INCLUDE_GenericCompactStack
/*------------------------------------------------------------------------*/

typedef struct GenericCompactStack GenericCompactStack;

struct GenericCompactStack
{
  void *base;			/* arbitrary (eg 64 bit) pointer */
  unsigned short ldsize;	/* at least 6 bit for 64 bit machines */
  unsigned short large_entries;	/* would actually only need 1 bit */
  unsigned count;		/* 32 bit wide => 16 GB size limit */
};

/*------------------------------------------------------------------------*/
#endif /* COMPSAT_INCLUDE_GenericCompactStack */
/*------------------------------------------------------------------------*/
/* The default pointer stack uses tree pointers and provides really fast
 * access to the stack contents.  It is actually very similar to the code
 * you would get when using a vector from the C++ STL.  Some stacks always
 * use this code and therefore we include the pointer stack code
 * unconditionally.
 */
typedef struct PointerStack PointerStack;

/*------------------------------------------------------------------------*/

struct PointerStack
{
  int *start, *top, *end;
};

/*------------------------------------------------------------------------*/

struct Queue
{
  int *start, *head, *tail, *end;
};

/*------------------------------------------------------------------------*/

struct Counter
{
  int current, init, inc;
  double factor, finished;
};

/*------------------------------------------------------------------------*/

struct Counters
{
  Counter restart;		/* counter for next restart */
  Counter rescore;		/* counter for next rescore */
  Counter report;		/* counter for next report */
};

/*------------------------------------------------------------------------*/

struct Statistics
{
  size_t max_bytes;		/* maximal number of allocated bytes */
  size_t bytes;			/* number of currently allocated bytes */
  double time;			/* accumulative seconds spent in library */
  double decisions;
  double conflicts;
  double assignments;
  unsigned reported;
  int variables;		/* number of occuring variables */
  int clauses;
  int literals;
};

/*------------------------------------------------------------------------*/

struct Limits
{
  double bytes;			/* byte limit (default -1 unrestricted) */
  double decisions;		/* decisions+1 (default -1 unrestricted) */
  double time;			/* time limit (default -1 unrestricted) */
};

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/
/* If checking is enabled we can check that a conflict clause to be added is
 * actually implied by the current clause set.  This check is performed by
 * unit resolution with a seperate simple 'Checker' solver.
 */
typedef struct Checker Checker;

struct Checker
{
  int enabled;			/* dynamically */
  CompSat *compsat;
  FastStack assignment;		/* our own assignment */
  FastStack *occurence[2];	/* literal occurence in compsat->literals */
  FastStack stack;		/* the pending assignment DFS stack */
  FastStack units;		/* the unit clauses */
};

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifndef COMPSAT_SLICE_VARIABLE
/*------------------------------------------------------------------------*/

struct Variable
{
  CompactStack watched[2];
  int assignment;		/* assignment, non zero -> assigned */
  int level;			/* level where this variable was assigned */
  int reason;			/* reason for this assignment */
  int pos[2];			/* position of literals in 'order' */
  int ref[2];			/* number of literals in clauses */
  int score[2];			/* score for each literal */
};

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

struct CompSat
{
  int level;
  CompactStack literals;	/* contains all literals and clauses */
  FastStack units;		/* indices of unit clauses */

#ifdef COMPSAT_SLICE_VARIABLE
  FastStack *watched[2];	/* the watched lists for each literal */
  FastStack assignment;		/* contains assignments for each variable */
  FastStack levels;		/* decision level of assigned variable */
  FastStack reason;		/* reason for assignments */
  FastStack pos[2];		/* position in order */
  FastStack ref[2];		/* the number of occurences */
  FastStack score[2];		/* score for each literal */
#else
  Variable *variables;
#endif

  FastStack mark[2];		/* bit map for mark flags of literals */
  FastStack used;		/* bit map of used variables */

  Queue queue;			/* the pending assignment BFS queue */
  FastStack trail;		/* the indices of assigned variables */
  FastStack control;		/* maps decision to trail level */
  FastStack order;		/* orders the literals */
  FastStack clause;		/* for conflicts and simplifying */
  int unassigned;		/* first potentially unassigned variable */
  int zero;			/* first necessarilly zero score variable */
  FastStack visited;		/* save visited literals */
  FastStack stack;		/* multipurpose stack */
  int max;			/* maximal variable id */
  int size;			/* size of watched arrays or variables */
  int contains_empty_clause;
  int restart_at_next_decision;

  FILE *log;			/* log file */
  FILE *out;			/* output file */
  int divert;			/* divert logging to output */
  int verbosity;		/* verbosity level for logging */
  int no_output;		/* disable all output (except logging) */

  char *error;			/* contains error message */
  char *name;			/* file name */

  int internal;			/* number of nested internal invocations */
  double entered;		/* enter time */

  Limits limits;
  Statistics stats;
  Counters counters;

#ifdef COMPSAT_CHECK
  size_t *leaked_ptr;		/* report leaked bytes */
  Checker checker;		/* e.g. conflict clauses implication */
#endif
#ifdef COMPSAT_WHITE
  int force_out_of_memory;	/* see 'memout' white box test */
#endif
};

/*------------------------------------------------------------------------*/

struct COMPSAT_ArgvIteratorInternal
{
  CompSat *compsat;
  int argc;			/* size of 'argv' ('argv[0]' is skipped) */
  char **argv;			/* copy of original 'argv' argument */
  int pos;			/* current position in 'argv' */
  char opt_buffer[3];		/* holds '-t' if '-t1.2' specified etc. */
};

/*------------------------------------------------------------------------*/

inline_static void *
new (CompSat * compsat, size_t size)
{
  size_t *res, bytes;

  if (!size)
    return 0;

  bytes = size;
#ifdef COMPSAT_CHECK
  bytes += sizeof (size_t);
#endif

#ifdef COMPSAT_WHITE
  if (compsat->force_out_of_memory)
    res = 0;
  else
#endif
    res = (size_t *) malloc (bytes);

  if (!res)
    return 0;

#ifdef COMPSAT_CHECK
  *res++ = size;
#endif
  memset (res, 0, size);
  compsat->stats.bytes += size;
  if (compsat->stats.max_bytes < compsat->stats.bytes)
    compsat->stats.max_bytes = compsat->stats.bytes;

  return (void *) res;
}

/*------------------------------------------------------------------------*/

inline_static void
delete (CompSat * compsat, void *ptr, size_t size)
{
  size_t *data;

  if (!ptr)
    {
      assert (!size);
      return;
    }

#ifdef COMPSAT_CHECK
  data = ((size_t *) ptr) - 1;
  assert (*data == size);
#else
  data = ptr;
  (void) size;
#endif
  assert (compsat->stats.bytes >= size);
  compsat->stats.bytes -= size;
  free ((void *) data);
}

/*-----------------------------------------------------------------------*/

inline_static void *
resize (CompSat * compsat, void *ptr, size_t old_size, size_t new_size)
{
  size_t *data, bytes;

  if (!ptr)
    {
      assert (!old_size);
      return new (compsat, new_size);
    }

  if (!new_size)
    {
      delete (compsat, ptr, old_size);
      return 0;
    }

#ifdef COMPSAT_CHECK
  bytes = new_size + sizeof (size_t);
  data = ((size_t *) ptr) - 1;
  assert (*data == old_size);
#else
  bytes = new_size;
  data = ptr;
#endif

#ifdef COMPSAT_WHITE
  if (compsat->force_out_of_memory)
    data = 0;
  else
#endif
    data = realloc (data, bytes);

  if (!data)
    return 0;

#ifdef COMPSAT_CHECK
  *data++ = new_size;
#endif

  if (new_size > old_size)
    memset (old_size + (char *) data, 0, new_size - old_size);

  assert (compsat->stats.bytes >= old_size);
  compsat->stats.bytes -= old_size;
  compsat->stats.bytes += new_size;
  if (compsat->stats.max_bytes < compsat->stats.bytes)
    compsat->stats.max_bytes = compsat->stats.bytes;

  return data;
}

/*------------------------------------------------------------------------*/

inline_static char *
copy_string (CompSat * compsat, const char *str)
{
  char *res;

  res = 0;

  if (str)
    {
      res = (char *) new (compsat, strlen (str) + 1);
      if (res)
	strcpy (res, str);
    }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
delete_string (CompSat * compsat, char *str)
{
  if (str)
    delete (compsat, str, strlen (str) + 1);
}

/*------------------------------------------------------------------------*/

double
COMPSAT_get_time (void)
{
  struct rusage u;
  double res;

  res = 0;

  if (!getrusage (RUSAGE_SELF, &u))
    {
      res += u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
      res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
    }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static double
delta_time (CompSat * compsat)
{
  double res;

  assert (compsat->internal);

  res = COMPSAT_get_time () - compsat->entered;
  if (res <= 0)
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
update_time (CompSat * compsat)
{
  double current, delta;

  assert (compsat->internal);

  current = COMPSAT_get_time ();
  delta = current - compsat->entered;
  compsat->stats.time += delta;
  compsat->entered = current;
}

/*------------------------------------------------------------------------*/

inline_static void
init_Counter (Counter * counter, int init, double factor, int inc)
{
  assert (init >= 0);

  counter->factor = factor;
  counter->init = init;
  counter->inc = inc;
  counter->current = init;
  counter->finished = 0;
}

/*------------------------------------------------------------------------*/

inline_static int
done_Counter (Counter * counter)
{
  int res;

  if (counter->init)
    {
      if (counter->current)
	{
	  assert (counter->current > 0);
	  counter->current -= 1;
	  res = 0;
	}
      else
	{
	  counter->init *= counter->factor;
	  counter->init += counter->inc;
	  counter->current = counter->init;
	  counter->finished++;
	  res = 1;
	}
    }
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_INCLUDE_CompactStack32
/*------------------------------------------------------------------------*/

inline_static int
ld_size_CompactStack32 (CompactStack32 * stack)
{
  int res;

  res = stack->word[0];
  assert (res);
  res &= 3;
  res |= (stack->word[1] >> 27) & ~3;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
size_CompactStack32 (CompactStack32 * stack)
{
  if (stack->word[0])
    return 1 << ld_size_CompactStack32 (stack);
  else
    return 0;
}

/*------------------------------------------------------------------------*/

inline_static void
double_size_CompactStack32 (CompactStack32 * stack)
{
#ifndef NDEBUG
  int old_size = size_CompactStack32 (stack);
#endif
  unsigned w0;

  /* The number of valid bits for the counter is restricted to be 28, thus
   * we can only allocate 2^28 32 bit words which results in 1 GB.
   */
  assert (ld_size_CompactStack32 (stack) < 28);

  w0 = stack->word[0] + 1;	/* increment lower 2 bits of 'ldsize' */

  /* carry has to go into word[1]
   */
  if (!(w0 & 3))
    {
      w0 -= 4;
      stack->word[1] += 0x20000000;
    }

  stack->word[0] = w0;

  assert (2 * old_size == size_CompactStack32 (stack));
}

/*------------------------------------------------------------------------*/

inline_static int
has_large_entries_CompactStack32 (CompactStack32 * stack)
{
  return (stack->word[1] & 0x10000000) != 0;
}

/*------------------------------------------------------------------------*/

inline_static size_t
bytes_CompactStack32 (CompactStack32 * stack)
{
  size_t res;

  res = size_CompactStack32 (stack);
  if (res)
    {
      res *= 2;
      if (has_large_entries_CompactStack32 (stack))
	res *= 2;
    }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
count_CompactStack32 (CompactStack32 * stack)
{
  return stack->word[1] & 0x0fffffff;
}

/*------------------------------------------------------------------------*/

inline_static void
reset_CompactStack32 (CompSat * compsat, CompactStack32 * stack, int count)
{
  unsigned w1, large_entries;

  (void) compsat;

  assert (count >= 0);
  assert (count <= count_CompactStack32 (stack));
  w1 = stack->word[1];
  large_entries = (w1 & 0x10000000) != 0;
  assert (large_entries == has_large_entries_CompactStack32 (stack));
  w1 &= count ? 0xf0000000 : 0xe0000000;
  w1 |= count;
  stack->word[1] = w1;

  if (!count && large_entries && stack->word[0])
    {
      /* If the stack becomes empty and has been using 4 bytes entries, then
       * we set the entry size to 2 bytes again.  In this case we have
       * to double the size.
       */
      double_size_CompactStack32 (stack);
    }

  assert (count_CompactStack32 (stack) == count);
}

/*------------------------------------------------------------------------*/

inline_static void
release_CompactStack32 (CompSat * compsat, CompactStack32 * stack)
{
  delete (compsat,
	  (void *) (stack->word[0] & ~3), bytes_CompactStack32 (stack));
  memset (stack, 0, sizeof (*stack));
}

/*------------------------------------------------------------------------*/

inline_static void
enlarge_CompactStack32 (CompSat * compsat, CompactStack32 * stack)
{
  size_t old_bytes, new_bytes;
  unsigned w0;
  void *ptr;

  w0 = stack->word[0];

  if (w0)
    {
      old_bytes = bytes_CompactStack32 (stack);
      double_size_CompactStack32 (stack);
      new_bytes = bytes_CompactStack32 (stack);
      ptr = (void *) (w0 & ~3);
      ptr = resize (compsat, ptr, old_bytes, new_bytes);
      stack->word[0] = ((w0 + 1) & 3) | (unsigned) ptr;
    }
  else
    stack->word[0] = (unsigned) new (compsat, 2);
}

/*------------------------------------------------------------------------*/

inline_static void
prepare_for_large_entries_CompactStack32 (CompSat * compsat,
					  CompactStack32 * stack)
{
  short *src, *p, *end;
  int *dst, *q;
  size_t bytes;
  unsigned w0;

  assert (!has_large_entries_CompactStack32 (stack));
  w0 = stack->word[0];
  assert (w0);
  p = src = (void *) (w0 & ~3);
  end = src + count_CompactStack32 (stack);
  bytes = 2 * size_CompactStack32 (stack);
  q = dst = new (compsat, 2 * bytes);
  while (p < end)
    *q++ = *p++;
  delete (compsat, src, bytes);
  stack->word[0] = (w0 & 3) | (unsigned) dst;
  stack->word[1] |= 0x10000000;
}

/*------------------------------------------------------------------------*/

inline_static void *
start_CompactStack32 (CompactStack32 * stack)
{
  return (void *) (stack->word[0] & ~3);
}

/*------------------------------------------------------------------------*/

inline_static void *
top_CompactStack32 (CompactStack32 * stack)
{
  void *start, *res;
  int count;

  count = count_CompactStack32 (stack);
  start = start_CompactStack32 (stack);

  if (has_large_entries_CompactStack32 (stack))
    res = count + (int *) start;
  else
    res = count + (short *) start;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
next_CompactStack32_aux (char **cursor_ptr, const int large,
			 const ptrdiff_t inc)
{
  char *cursor;
  int res;

  cursor = *cursor_ptr;
  res = large ? *(int *) cursor : *(short *) cursor;
  *cursor_ptr = cursor + inc;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
push_CompactStack32 (CompSat * compsat, CompactStack32 * stack, int elem)
{
  int *iptr, has_large;
  short *sptr;

  if (size_CompactStack32 (stack) <= count_CompactStack32 (stack))
    enlarge_CompactStack32 (compsat, stack);

  assert (size_CompactStack32 (stack) > count_CompactStack32 (stack));

  has_large = has_large_entries_CompactStack32 (stack);

  if (!has_large && (elem < -32768 || elem >= 32768))
    {
      has_large = 1;
      prepare_for_large_entries_CompactStack32 (compsat, stack);
    }

  if (has_large)
    {
      iptr = top_CompactStack32 (stack);
      *iptr = elem;
    }
  else
    {
      sptr = top_CompactStack32 (stack);
      *sptr = (int) elem;
    }

  stack->word[1]++;
}

/*------------------------------------------------------------------------*/

inline_static int
get_CompactStack32 (CompactStack32 * stack, int idx)
{
  int *iptr, has_large, res;
  void *start;
  short *sptr;

  if (count_CompactStack32 (stack) <= idx)
    return 0;

  has_large = has_large_entries_CompactStack32 (stack);
  start = start_CompactStack32 (stack);

  if (has_large)
    {
      iptr = start;
      iptr += idx;
      res = *iptr;
    }
  else
    {
      sptr = start;
      sptr += idx;
      res = *sptr;
    }

  return res;
}

/*------------------------------------------------------------------------*/

#define IteratorCompactStack32(stack, it, dir) \
  char * it ## _start = start_CompactStack32 (stack); \
  char * it ## _top = top_CompactStack32 (stack); \
  const int it ## _large = has_large_entries_CompactStack32 (stack); \
  const ptrdiff_t it ## _inc = (dir) * (it ## _large ? 4 : 2); \
  const char * it ## _done = \
    ((dir) < 0) ? it ## _start + it ## _inc : it ## _top; \
  char * it ## _cursor = \
    (((dir) < 0) ? it ## _top + it ## _inc : it ## _start)

/*------------------------------------------------------------------------*/

#define CursorCompactStack32(stack, it, pos, dir) \
  char * it ## _start = start_CompactStack32 (stack); \
  const int it ## _large = has_large_entries_CompactStack32 (stack); \
  const ptrdiff_t it ## _inc = (dir) * (it ## _large ? 4 : 2); \
  char * it ## _cursor = it ## _start + (pos) * (it ## _large ? 4 : 2)

/*------------------------------------------------------------------------*/

#define save_CursorCompactStack32(it, saved_ptr) \
  do { (saved_ptr) = it ## _cursor; } while(0)

/*------------------------------------------------------------------------*/

#define restore_CursorCompactStack32(stack, saved_ptr) \
  do { it ## _cursor = (saved_ptr); } while(0)

/*------------------------------------------------------------------------*/

#define swap_CompactStack32(stack, it, a, b) \
do { \
  int it ## _tmp; \
  assert (it ## _start <= (char*)(a)); \
  assert (it ## _start <= (char*)(b)); \
  assert ((a) < top_CompactStack32(stack)); \
  assert ((b) < top_CompactStack32(stack)); \
  assert (it ## _large == has_large_entries_CompactStack32 (stack)); \
  if(it ## _large) \
    { \
      it ## _tmp = *(int*)(a); \
      *(int*)(a) = *(int*)(b); \
      *(int*)(b) = it ## _tmp; \
    } \
  else \
    { \
      it ## _tmp = *(short*)(a); \
      *(short*)(a) = *(short*)(b); \
      *(short*)(b) = it ## _tmp; \
    } \
} while(0)

/*------------------------------------------------------------------------*/

#define done_CompactStack32(it) \
  (it ## _cursor == it ## _done)

/*------------------------------------------------------------------------*/

#define next_CompactStack32(it) \
  next_CompactStack32_aux(&it ## _cursor, it ## _large, it ## _inc)

/*------------------------------------------------------------------------*/
#endif /* COMPSAT_INCLUDE_CompactStack32 */
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_INCLUDE_GenericCompactStack
/*------------------------------------------------------------------------*/

inline_static int
count_GenericCompactStack (GenericCompactStack * stack)
{
  return stack->count;
}

/*------------------------------------------------------------------------*/

inline_static int
size_GenericCompactStack (GenericCompactStack * stack)
{
  if (stack->base)
    return 1 << stack->ldsize;
  else
    return 0;
}

/*------------------------------------------------------------------------*/

inline_static size_t
bytes_GenericCompactStack (GenericCompactStack * stack)
{
  size_t res;

  res = 2 * (size_t) size_GenericCompactStack (stack);
  if (stack->large_entries)
    res *= 2;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
reset_GenericCompactStack (CompSat * compsat, GenericCompactStack * stack,
			   int count)
{
  assert (count >= 0);
  assert (count <= stack->count);

  if (!count && stack->large_entries)
    {
      stack->large_entries = 0;
      stack->ldsize += 1;
    }

  stack->count = count;
}

/*------------------------------------------------------------------------*/

inline_static void
release_GenericCompactStack (CompSat * compsat, GenericCompactStack * stack)
{
  delete (compsat, stack->base, bytes_GenericCompactStack (stack));
  memset (stack, 0, sizeof (*stack));
}

/*------------------------------------------------------------------------*/

inline_static void
enlarge_GenericCompactStack (CompSat * compsat, GenericCompactStack * stack)
{
  size_t old_bytes, new_bytes;

  if (stack->base)
    {
      old_bytes = bytes_GenericCompactStack (stack);
      stack->ldsize++;
      new_bytes = bytes_GenericCompactStack (stack);
      stack->base = resize (compsat, stack->base, old_bytes, new_bytes);
    }
  else
    {
      assert (!stack->large_entries);
      assert (!stack->ldsize);
      stack->base = new (compsat, 2);
    }
}

/*------------------------------------------------------------------------*/

inline_static void
prepare_for_large_entries_GenericCompactStack (CompSat * compsat,
					       GenericCompactStack * stack)
{
  short *src, *p, *end;
  int *dst, *q;
  size_t bytes;

  assert (!stack->large_entries);
  bytes = bytes_GenericCompactStack (stack);
  q = dst = (int *) new (compsat, 2 * bytes);
  p = src = (short *) stack->base;
  end = src + size_GenericCompactStack (stack);
  while (p < end)
    *q++ = *p++;
  delete (compsat, src, bytes);
  stack->large_entries = 1;
  stack->base = dst;
}

/*------------------------------------------------------------------------*/

inline_static void *
top_GenericCompactStack (GenericCompactStack * stack)
{
  short *res;
  int count;

  count = stack->count;
  if (stack->large_entries)
    count *= 2;

  res = stack->base;
  res += count;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
next_GenericCompactStack_aux (char **cursor_ptr, const int large,
			      const ptrdiff_t inc)
{
  char *cursor;
  int res;

  cursor = *cursor_ptr;
  res = large ? *(int *) cursor : *(short *) cursor;
  *cursor_ptr = cursor + inc;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
push_GenericCompactStack (CompSat * compsat,
			  GenericCompactStack * stack, int elem)
{
  int *iptr, has_large;
  short *sptr;

  if (size_GenericCompactStack (stack) <= stack->count)
    enlarge_GenericCompactStack (compsat, stack);

  assert (size_GenericCompactStack (stack) > stack->count);

  has_large = stack->large_entries;

  if (!has_large && (elem < -32768 || elem >= 32768))
    {
      has_large = 1;
      prepare_for_large_entries_GenericCompactStack (compsat, stack);
    }

  if (has_large)
    {
      iptr = top_GenericCompactStack (stack);
      *iptr = elem;
    }
  else
    {
      sptr = top_GenericCompactStack (stack);
      *sptr = (int) elem;
    }

  stack->count++;
}

/*------------------------------------------------------------------------*/

inline_static int
get_GenericCompactStack (GenericCompactStack * stack, int idx)
{
  int *iptr, has_large, res;
  void *start;
  short *sptr;

  if (stack->count <= idx)
    return 0;

  has_large = stack->large_entries;
  start = stack->base;

  if (has_large)
    {
      iptr = start;
      iptr += idx;
      res = *iptr;
    }
  else
    {
      sptr = start;
      sptr += idx;
      res = *sptr;
    }

  return res;
}

/*------------------------------------------------------------------------*/

#define IteratorGenericCompactStack(stack, it, dir) \
  char * it ## _start = (stack)->base; \
  char * it ## _top = top_GenericCompactStack (stack); \
  const int it ## _large = (stack)->large_entries; \
  const ptrdiff_t it ## _inc = (dir) * (it ## _large ? 4 : 2); \
  const char * it ## _done = \
    ((dir) < 0) ? it ## _start + it ## _inc : it ## _top; \
  char * it ## _cursor = \
    (((dir) < 0) ? it ## _top + it ## _inc : it ## _start)

/*------------------------------------------------------------------------*/

#define CursorGenericCompactStack(stack, it, pos, dir) \
  char * it ## _start = (stack)->base; \
  const int it ## _large = (stack)->large_entries; \
  const ptrdiff_t it ## _inc = (dir) * (it ## _large ? 4 : 2); \
  char * it ## _cursor = it ## _start + (pos) * (it ## _large ? 4 : 2)

/*------------------------------------------------------------------------*/

#define save_CursorGenericCompactStack(it, saved_ptr) \
  do { (saved_ptr) = it ## _cursor; } while(0)

/*------------------------------------------------------------------------*/

#define restore_CursorGenericCompactStack(stack, saved_ptr) \
  do { it ## _cursor = (saved_ptr); } while(0)

/*------------------------------------------------------------------------*/

#define swap_GenericCompactStack(stack, it, a, b) \
do { \
  int it ## _tmp; \
  assert (it ## _start <= (char*)(a)); \
  assert (it ## _start <= (char*)(b)); \
  assert ((a) < top_GenericCompactStack(stack)); \
  assert ((b) < top_GenericCompactStack(stack)); \
  assert (it ## _large == (stack)->large_entries); \
  if(it ## _large) \
    { \
      it ## _tmp = *(int*)(a); \
      *(int*)(a) = *(int*)(b); \
      *(int*)(b) = it ## _tmp; \
    } \
  else \
    { \
      it ## _tmp = *(short*)(a); \
      *(short*)(a) = *(short*)(b); \
      *(short*)(b) = it ## _tmp; \
    } \
} while(0)

/*------------------------------------------------------------------------*/

#define done_GenericCompactStack(it) \
  (it ## _cursor == it ## _done)

/*------------------------------------------------------------------------*/

#define next_GenericCompactStack(it) \
  next_GenericCompactStack_aux(&it ## _cursor, it ## _large, it ## _inc)

/*------------------------------------------------------------------------*/
#endif /* COMPSAT_INCLUDE_GenericCompactStack */
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/

inline_static void
check_PointerStack (PointerStack * stack)
{
  assert (stack->start <= stack->top);
  assert (stack->top <= stack->end);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static size_t
bytes_PointerStack (PointerStack * stack)
{
  return ((char *) stack->end) - ((char *) stack->start);
}

/*------------------------------------------------------------------------*/

inline_static int
size_PointerStack (PointerStack * stack)
{
  return stack->end - stack->start;
}

/*------------------------------------------------------------------------*/

inline_static int
count_PointerStack (PointerStack * stack)
{
  return stack->top - stack->start;
}

/*------------------------------------------------------------------------*/

inline_static void
reset_PointerStack (CompSat * compsat, PointerStack * stack, int count)
{
  assert (count >= 0);
  assert (count <= count_PointerStack (stack));
  stack->top = stack->start + count;
#ifdef COMPSAT_CHECK
  check_PointerStack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
release_PointerStack (CompSat * compsat, PointerStack * stack)
{
  delete (compsat, stack->start, bytes_PointerStack (stack));
  memset (stack, 0, sizeof (*stack));
#ifdef COMPSAT_CHECK
  check_PointerStack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
enlarge_PointerStack (CompSat * compsat, PointerStack * stack)
{
  size_t old_bytes, new_bytes;
  int *old_start, *new_start;
  ptrdiff_t delta;

  old_bytes = bytes_PointerStack (stack);
  new_bytes = old_bytes ? 2 * old_bytes : sizeof (int);

  old_start = stack->start;
  new_start = (int *) resize (compsat, old_start, old_bytes, new_bytes);
  delta = new_start - old_start;

  stack->start = new_start;
  stack->top += delta;
  stack->end = (int *) (new_bytes + (char *) stack->start);

#ifdef COMPSAT_CHECK
  check_PointerStack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
push_PointerStack (CompSat * compsat, PointerStack * stack, int n)
{
  assert (n != INT_MIN);

  if (stack->top == stack->end)
    enlarge_PointerStack (compsat, stack);

  *stack->top++ = n;

#ifdef COMPSAT_CHECK
  check_PointerStack (stack);
#endif
}

/*------------------------------------------------------------------------*/
#if defined(COMPSAT_CHECK) || defined(COMPSAT_WHITE)
/*------------------------------------------------------------------------*/

inline_static int
is_empty_PointerStack (PointerStack * stack)
{
  return stack->top == stack->start;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
pop_PointerStack (CompSat * compsat, PointerStack * stack)
{
  int res;

  assert (!is_empty_PointerStack (stack));
  res = *--stack->top;
#ifdef COMPSAT_CHECK
  check_PointerStack (stack);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
check_legal_index (int idx)
{
  assert (idx >= 0);
  assert (idx != FALSE);
  assert (idx != TRUE);
}

/*------------------------------------------------------------------------*/

inline_static void
set_PointerStack (CompSat * compsat, PointerStack * stack, int idx, int val)
{
  int *p;

  check_legal_index (idx);

  while (size_PointerStack (stack) <= idx)
    enlarge_PointerStack (compsat, stack);

  p = stack->start + idx;
  if (stack->top <= p)
    {
      memset (stack->top, 0, (char *) p - (char *) stack->top);
      stack->top = p + 1;
    }

  *p = val;

#ifdef COMPSAT_CHECK
  check_PointerStack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
get_PointerStack (PointerStack * stack, int idx)
{
  int *p, res;

  check_legal_index (idx);
  p = stack->start + idx;
  if (p >= stack->top)
    res = 0;
  else
    res = *p;

  return res;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_SLICE_VARIABLE
/*------------------------------------------------------------------------*/

inline_static void
inc_PointerStack (CompSat * compsat, PointerStack * stack, int idx)
{
  int word;

  word = get_PointerStack (stack, idx);
  word++;
  set_PointerStack (compsat, stack, idx, word);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static int
next_PointerStack_aux (int **ptr, const int delta)
{
  int res;

  res = **ptr;
  *ptr += delta;

  return res;
}

/*------------------------------------------------------------------------*/

#define IteratorPointerStack(stack, it, dir) \
  int * it ## _start = (stack)->start; \
  int * it ## _top = (stack)->top; \
  const int it ## _inc = (dir); \
  const int * it ## _done = \
    ((dir) < 0) ? it ## _start - 1 : it ## _top; \
  int * it ## _cursor = \
    ((dir) < 0) ? it ## _top - 1 : it ## _start

/*------------------------------------------------------------------------*/

#define CursorPointerStack(stack, it, pos, dir) \
  int * it ## _start = (stack)->start; \
  const int it ## _inc = (dir); \
  int * it ## _cursor = it ## _start + (pos)

/*------------------------------------------------------------------------*/

#define save_CursorPointerStack(it, saved_ptr) \
do { (saved_ptr) = it ## _cursor; } while(0)

/*------------------------------------------------------------------------*/

#define restore_CursorPointerStack(stack, saved_ptr) \
  do { it ## _cursor = (saved_ptr); } while(0)

/*------------------------------------------------------------------------*/

#define swap_PointerStack(stack, it, a, b) \
do { \
  int it ## _tmp; \
  assert ((stack)->start <= (int*)(a) && (int*)(a) < (stack)->top); \
  assert ((stack)->start <= (int*)(b) && (int*)(b) < (stack)->top); \
  it ## _tmp = *(int*)(a); \
  *(int*)(a) = *(int*)(b); \
  *(int*)(b) = it ## _tmp; \
} while(0)

/*------------------------------------------------------------------------*/

#define done_PointerStack(it) \
  (it ## _cursor == it ## _done)

/*------------------------------------------------------------------------*/

#define next_PointerStack(it) \
  next_PointerStack_aux (&it ## _cursor, it ## _inc)

/*------------------------------------------------------------------------*/
#ifndef COMPSAT_SLICE_VARIABLE
/*------------------------------------------------------------------------*/

inline_static void
release_Variable (CompSat * compsat, Variable * v)
{
  release_CompactStack (compsat, &v->watched[0]);
  release_CompactStack (compsat, &v->watched[1]);
}

/*------------------------------------------------------------------------*/

inline_static void
release_variables (CompSat * compsat)
{
  int i;

  for (i = 1; i <= compsat->max; i++)
    release_Variable (compsat, compsat->variables + i);

  delete (compsat, compsat->variables, compsat->size * sizeof (Variable));
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/

inline_static void
check_Queue (Queue * queue)
{
  assert (queue->start <= queue->head);
  assert (queue->start <= queue->tail);

  if (queue->start)
    {
      assert (queue->head < queue->end);
      assert (queue->tail < queue->end);
    }
  else
    {
      assert (!queue->head);
      assert (!queue->tail);
    }
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
reset_Queue (CompSat * compsat, Queue * queue)
{
  queue->head = queue->tail = queue->start;
#ifdef COMPSAT_CHECK
  check_Queue (queue);
#endif
}

/*------------------------------------------------------------------------*/

inline_static size_t
size_Queue (Queue * queue)
{
  return queue->end - queue->start;
}

/*------------------------------------------------------------------------*/

inline_static size_t
bytes_Queue (Queue * queue)
{
  return size_Queue (queue) * sizeof (queue->start[0]);
}

/*------------------------------------------------------------------------*/

inline_static void
release_Queue (CompSat * compsat, Queue * queue)
{
  delete (compsat, queue->start, bytes_Queue (queue));
  memset (queue, 0, sizeof (Queue));
#ifdef COMPSAT_CHECK
  check_Queue (queue);
#endif
}

/*------------------------------------------------------------------------*/

inline_static FILE *
out_file (CompSat * compsat)
{
  FILE *file;

  file = compsat->out;
  if (!file)
    file = stdout;

  return file;
}

/*------------------------------------------------------------------------*/

inline_static FILE *
log_file (CompSat * compsat)
{
  FILE *file;

  if (compsat->divert)
    return out_file (compsat);

  file = compsat->log;
  if (!file)
    file = stdout;

  return file;
}

/*------------------------------------------------------------------------*/

inline_static void
log (CompSat * compsat, int level, const char *fmt, ...)
{
  FILE *file;
  va_list ap;

#ifdef COMPSAT_CHECK
  char *p;
  p = strchr (fmt, '\n');
  assert (p);
  assert (!strchr (p + 1, '\n'));
#endif

  if (level > compsat->verbosity)
    return;

  file = log_file (compsat);
  fputs (COMPSAT_PREFIX, file);
  va_start (ap, fmt);
  vfprintf (file, fmt, ap);
  va_end (ap);
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

inline_static void
out (CompSat * compsat, const char *fmt, ...)
{
  FILE *file;
  va_list ap;

  if (compsat->no_output)
    return;

  file = out_file (compsat);
  va_start (ap, fmt);
  vfprintf (file, fmt, ap);
  va_end (ap);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
ABS (int lit)
{
  return (lit < 0) ? -lit : lit;
}

/*------------------------------------------------------------------------*/

inline_static int
SGN (int lit)
{
  return (lit < 0);
}

/*------------------------------------------------------------------------*/

inline_static int
unit2assignment (CompSat * compsat, int lit)
{
  assert (0 < ABS (lit));
  assert (ABS (lit) <= compsat->max);
  return lit - MAX_VAR;
}

/*------------------------------------------------------------------------*/

inline_static void
split_assignment (CompSat * compsat, int assignment, int *lit_ptr,
		  int *reason_ptr)
{
  int lit, reason;

  if (assignment <= 0)
    {
      lit = assignment + MAX_VAR;
      reason = 0;
    }
  else
    {
      lit = get_CompactStack (&compsat->literals, assignment);
      reason = assignment;
    }

  assert (0 < ABS (lit));
  assert (ABS (lit) <= compsat->max);
  assert (0 <= reason);
  assert (reason < count_CompactStack (&compsat->literals));

  *reason_ptr = reason;
  *lit_ptr = lit;
}

/*------------------------------------------------------------------------*/

inline_static int
count_Queue (Queue * queue)
{
  int res;

  res = queue->tail - queue->head;
  if (res < 0)
    res += queue->end - queue->start;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
is_empty_Queue (Queue * queue)
{
  return !count_Queue (queue);
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_LOG
/*------------------------------------------------------------------------*/

inline_static void
log_assignment (CompSat * compsat, int level, const char *msg, int assignment)
{
  int reason, lit;

  if (compsat->verbosity >= level)
    {
      split_assignment (compsat, assignment, &lit, &reason);
      log (compsat, level, "%d %s %d %d\n", compsat->level, msg, lit, reason);
    }
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
enlarge_Queue (CompSat * compsat, Queue * queue)
{
  int old_bytes, new_bytes, *new_start, count;
  size_t n, m;

  old_bytes = bytes_Queue (queue);
  if (old_bytes)
    new_bytes = old_bytes * 2;
  else
    new_bytes = sizeof (queue->start[0]) * 2;

  new_start = (int *) new (compsat, new_bytes);

  if (queue->tail < queue->head)
    {
      n = ((char *) queue->end) - ((char *) queue->head);
      memcpy (new_start, queue->head, n);

      m = ((char *) queue->tail) - ((char *) queue->start);
      memcpy (n + (char *) new_start, queue->start, m);
    }
  else if (queue->tail > queue->head)
    {
      n = ((char *) queue->tail) - ((char *) queue->head);
      memcpy (new_start, queue->head, n);
    }

  count = count_Queue (queue);
  delete (compsat, queue->start, old_bytes);
  queue->start = new_start;
  queue->head = queue->start;
  queue->tail = queue->start + count;
  queue->end = (int *) (new_bytes + (char *) queue->start);

  assert (count_Queue (queue) + 1 < size_Queue (queue));
#ifdef COMPSAT_CHECK
  check_Queue (queue);
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
enqueue (CompSat * compsat, Queue * queue, int data)
{
  if (count_Queue (queue) + 1 >= size_Queue (queue))
    enlarge_Queue (compsat, queue);

  *queue->tail++ = data;
  if (queue->tail >= queue->end)
    queue->tail = queue->start;

  assert (!is_empty_Queue (queue));
#ifdef COMPSAT_CHECK
  check_Queue (queue);
#endif
#ifdef COMPSAT_LOG
  log_assignment (compsat, 5, "ENQUEUE", data);
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
dequeue (CompSat * compsat, Queue * queue)
{
  int res;

  assert (!is_empty_Queue (queue));

  res = *queue->head++;
  if (queue->head == queue->end)
    queue->head = queue->start;

#ifdef COMPSAT_CHECK
  check_Queue (queue);
#endif
#ifdef COMPSAT_LOG
  log_assignment (compsat, 5, "DEQUEUE", res);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
split_bit (int bit, int *word_ptr, int *mask_ptr)
{
  assert (sizeof (int) * 8 <= BITS_PER_WORD);
  check_legal_index (bit);
  *word_ptr = bit >> LD_BITS_PER_WORD;
  *mask_ptr = (1 << (bit & (BITS_PER_WORD - 1)));
}

/*------------------------------------------------------------------------*/

inline_static int
get_bit (CompSat * compsat, FastStack * stack, int bit)
{
  int mask, idx, word;

  split_bit (bit, &idx, &mask);
  word = get_FastStack (stack, idx);
  return (word & mask) != 0;
}

/*------------------------------------------------------------------------*/

inline_static void
clr_bit (CompSat * compsat, FastStack * stack, int bit)
{
  int mask, idx, word;

  split_bit (bit, &idx, &mask);
  word = get_FastStack (stack, idx);
  word &= ~mask;
  set_FastStack (compsat, stack, idx, word);
}

/*------------------------------------------------------------------------*/

inline_static void
set_bit (CompSat * compsat, FastStack * stack, int bit)
{
  int mask, idx, word;

  split_bit (bit, &idx, &mask);
  word = get_FastStack (stack, idx);
  word |= mask;
  set_FastStack (compsat, stack, idx, word);
}

/*------------------------------------------------------------------------*/

inline_static int
get_mark (CompSat * compsat, int lit)
{
  return get_bit (compsat, &compsat->mark[SGN (lit)], ABS (lit));
}

/*------------------------------------------------------------------------*/

inline_static void
set_mark (CompSat * compsat, int lit)
{
  set_bit (compsat, &compsat->mark[SGN (lit)], ABS (lit));
}

/*------------------------------------------------------------------------*/

inline_static void
clr_mark (CompSat * compsat, int lit)
{
  clr_bit (compsat, &compsat->mark[SGN (lit)], ABS (lit));
}

/*------------------------------------------------------------------------*/

inline_static int
get_used (CompSat * compsat, int idx)
{
  return get_bit (compsat, &compsat->used, idx);
}

/*------------------------------------------------------------------------*/

inline_static void
set_used (CompSat * compsat, int idx)
{
  set_bit (compsat, &compsat->used, idx);
}

/*------------------------------------------------------------------------*/

inline_static int
get_level (CompSat * compsat, int idx)
{
  assert (idx > 0);
#ifdef COMPSAT_SLICE_VARIABLE
  return get_FastStack (&compsat->levels, idx);
#else
  return compsat->variables[idx].level;
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
set_level (CompSat * compsat, int idx, int level)
{
  assert (idx > 0);
#ifdef COMPSAT_SLICE_VARIABLE
  set_FastStack (compsat, &compsat->levels, idx, level);
#else
  compsat->variables[idx].level = level;
#endif
}

/*------------------------------------------------------------------------*/

inline_static CompactStack *
get_watched (CompSat * compsat, int lit)
{
#ifdef COMPSAT_SLICE_VARIABLE
  return compsat->watched[SGN (lit)] + ABS (lit);
#else
  return &compsat->variables[ABS (lit)].watched[SGN (lit)];
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
inc_ref (CompSat * compsat, int lit)
{
#ifdef COMPSAT_SLICE_VARIABLE
  inc_FastStack (compsat, &compsat->ref[SGN (lit)], ABS (lit));
#else
  compsat->variables[ABS (lit)].ref[SGN (lit)]++;
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
get_ref (CompSat * compsat, int lit)
{
#ifdef COMPSAT_SLICE_VARIABLE
  return get_FastStack (&compsat->ref[SGN (lit)], ABS (lit));
#else
  return compsat->variables[ABS (lit)].ref[SGN (lit)];
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
get_assignment (CompSat * compsat, int idx)
{
  assert (idx > 0);
#ifdef COMPSAT_SLICE_VARIABLE
  return get_FastStack (&compsat->assignment, idx);
#else
  return compsat->variables[idx].assignment;
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
set_assignment (CompSat * compsat, int idx, int assignment)
{
  assert (idx > 0);
#ifdef COMPSAT_SLICE_VARIABLE
  set_FastStack (compsat, &compsat->assignment, idx, assignment);
#else
  compsat->variables[idx].assignment = assignment;
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
set_reason (CompSat * compsat, int idx, int reason)
{
  assert (idx > 0);
#ifdef COMPSAT_SLICE_VARIABLE
  set_FastStack (compsat, &compsat->reason, idx, reason);
#else
  compsat->variables[idx].reason = reason;
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
get_reason (CompSat * compsat, int idx)
{
  assert (idx > 0);
#ifdef COMPSAT_SLICE_VARIABLE
  return get_FastStack (&compsat->reason, idx);
#else
  return compsat->variables[idx].reason;
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
set_pos (CompSat * compsat, int lit, int pos)
{
#ifdef COMPSAT_SLICE_VARIABLE
  set_FastStack (compsat, &compsat->pos[SGN (lit)], ABS (lit), pos);
#else
  compsat->variables[ABS (lit)].pos[SGN (lit)] = pos;
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
get_pos (CompSat * compsat, int lit)
{
#ifdef COMPSAT_SLICE_VARIABLE
  return get_FastStack (&compsat->pos[SGN (lit)], ABS (lit));
#else
  return compsat->variables[ABS (lit)].pos[SGN (lit)];
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
get_score (CompSat * compsat, int lit)
{
#ifdef COMPSAT_SLICE_VARIABLE
  return get_FastStack (&compsat->score[SGN (lit)], ABS (lit));
#else
  return compsat->variables[ABS (lit)].score[SGN (lit)];
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
set_score (CompSat * compsat, int lit, int score)
{
#ifdef COMPSAT_SLICE_VARIABLE
  set_FastStack (compsat, &compsat->score[SGN (lit)], ABS (lit), score);
#else
  compsat->variables[ABS (lit)].score[SGN (lit)] = score;
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
add_score (CompSat * compsat, int lit, int delta)
{
  int old_score, new_score, pos;

  old_score = get_score (compsat, lit);
  new_score = old_score + delta;

  assert (new_score >= 0);

  if (!old_score && new_score > 0)
    {
      pos = get_pos (compsat, lit);
      if (compsat->zero >= pos)
	compsat->zero = pos + 1;
    }

  set_score (compsat, lit, new_score);
}

/*------------------------------------------------------------------------*/

inline_static int
copy_clause (CompSat * compsat, const int *clause)
{
  const int *p;
  int res;

  res = count_CompactStack (&compsat->literals);

  for (p = clause; *p; p++)
    push_CompactStack (compsat, &compsat->literals, *p);

  push_CompactStack (compsat, &compsat->literals, 0);

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
unmark_visited (CompSat * compsat)
{
  int *p;

  for (p = compsat->visited.start; p < compsat->visited.top; p++)
    clr_mark (compsat, *p);

  reset_FastStack (compsat, &compsat->visited, 0);
}

/*------------------------------------------------------------------------*/

inline_static int
simplify_clause (CompSat * compsat, const int *clause)
{
  int trivial, lit, res;
  const int *p;

  res = count_CompactStack (&compsat->literals);
  trivial = 0;
  p = clause;

  for (;;)
    {
      lit = *p;
      if (!lit)
	break;

      if ((trivial = get_mark (compsat, -lit)))
	break;

      p++;

      if (get_mark (compsat, lit))
	continue;

      set_mark (compsat, lit);
      push_FastStack (compsat, &compsat->visited, lit);

      push_CompactStack (compsat, &compsat->literals, lit);
    }

  if (trivial)
    {
      reset_CompactStack (compsat, &compsat->literals, res);
      res = 0;
    }
  else
    push_CompactStack (compsat, &compsat->literals, 0);

  unmark_visited (compsat);

  return res;
}

/*------------------------------------------------------------------------*/
#if defined(COMPSAT_SLICE_VARIABLE) || defined(COMPSAT_CHECK)
/*------------------------------------------------------------------------*/

inline_static void
enlarge_stack2 (CompSat * compsat, FastStack ** stack, int old_size,
		int new_size)
{
  int i, new_bytes, old_bytes;

  old_bytes = old_size * sizeof (FastStack);
  new_bytes = new_size * sizeof (FastStack);

  for (i = 0; i <= 1; i++)
    stack[i] = resize (compsat, stack[i], old_bytes, new_bytes);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
enlarge_aux (CompSat * compsat)
{
  int old_size, new_size;

  old_size = compsat->size;
  new_size = old_size ? 2 * old_size : 1;
#ifdef COMPSAT_SLICE_VARIABLE
  enlarge_stack2 (compsat, compsat->watched, old_size, new_size);
#else
  compsat->variables = (Variable *)
    resize (compsat, compsat->variables,
	    old_size * sizeof (Variable), new_size * sizeof (Variable));
#endif
#ifdef COMPSAT_CHECK
  enlarge_stack2 (compsat, compsat->checker.occurence, old_size, new_size);
#endif
  compsat->size = new_size;
}

/*------------------------------------------------------------------------*/

inline_static void
enlarge (CompSat * compsat, int idx)
{
  while (idx >= compsat->size)
    enlarge_aux (compsat);

  if (idx > compsat->max)
    {
      compsat->max = idx;
      assert (idx < MAX_VAR);
    }
}

/*------------------------------------------------------------------------*/

inline_static void
insert_variables (CompSat * compsat, const int *clause)
{
  const int *p;
  int idx, max;

  max = -1;

  for (p = clause; *p; p++)
    {
      if ((idx = ABS (*p)) > max)
	max = idx;

      if (!get_used (compsat, idx))
	{
	  compsat->stats.variables++;
	  set_used (compsat, idx);
	}
    }

  enlarge (compsat, max);
}

/*------------------------------------------------------------------------*/

inline_static void
order_clause (CompSat * compsat, int clause_idx)
{
  CursorCompactStack (&compsat->literals, it, clause_idx, 1);
  void *saved, *next, *watched, *start;
  int idx, level, max_level, count;

  save_CursorCompactStack (it, next);

  for (count = 0; count < MAX_WATCHED; count++)
    {
      saved = start = next;

      if (!(idx = ABS (next_CompactStack (it))))
	break;

      save_CursorCompactStack (it, next);

      max_level = -1;
      watched = 0;

      do
	{
	  level = get_level (compsat, idx);
	  if (level > max_level)
	    {
	      watched = saved;
	      max_level = level;
	    }

	  save_CursorCompactStack (it, saved);
	}
      while ((idx = ABS (next_CompactStack (it))));

      assert (watched);
      swap_CompactStack (&compsat->literals, it, start, watched);
      restore_CursorCompactStack (it, next);
    }
}

/*------------------------------------------------------------------------*/
#if defined(COMPSAT_SLICE_VARIABLE) || defined(COMPSAT_CHECK)
/*------------------------------------------------------------------------*/

inline_static void
release_stack2 (CompSat * compsat, FastStack ** stack, int size)
{
  int i, j;

  for (i = 0; i <= 1; i++)
    {
      for (j = 0; j < size; j++)
	release_FastStack (compsat, stack[i] + j);

      delete (compsat, stack[i], size * sizeof (FastStack));
    }
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/

inline_static int
is_connected (CompSat * compsat, int literal_idx)
{
  CompactStack *watched;
  int lit, tmp, res;

  lit = get_CompactStack (&compsat->literals, literal_idx);
  watched = get_watched (compsat, lit);

  {
    IteratorCompactStack (watched, it, 1);
    res = 0;
    while (!res && !done_CompactStack (it))
      {
	tmp = next_CompactStack (it);
	if (tmp == literal_idx)
	  res = 1;
      }
  }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
check_clause (CompSat * compsat, int clause_idx)
{
  CursorCompactStack (&compsat->literals, it, clause_idx - 1, 1);
  int i, tmp;

  tmp = next_CompactStack (it);
  assert (!tmp);

  for (i = 0; i < MAX_WATCHED && next_CompactStack (it); i++)
    assert (is_connected (compsat, clause_idx + i));
}

/*------------------------------------------------------------------------*/

inline_static void
check (CompSat * compsat)
{
  IteratorCompactStack (&compsat->literals, literals_it, 1);
  CompactStack *watched;
  int lit, i, s, tmp;

  tmp = next_CompactStack (literals_it);
  assert (!tmp);

  for (i = 1; !done_CompactStack (literals_it); i++)
    {
      check_clause (compsat, i);
      while (next_CompactStack (literals_it))
	i++;
    }

  for (i = 1; i <= compsat->max; i++)
    for (s = -1; s <= 1; s += 2)
      {
	lit = i * s;
	watched = get_watched (compsat, lit);

	{
	  IteratorCompactStack (watched, watched_it, 1);
	  while (!done_CompactStack (watched_it))
	    {
	      tmp = next_CompactStack (watched_it);
	      assert (get_CompactStack (&compsat->literals, tmp) == lit);
	    }
	}
      }
}

/*------------------------------------------------------------------------*/

inline_static void
init_Checker (Checker * checker, CompSat * compsat)
{
  checker->compsat = compsat;
}

/*------------------------------------------------------------------------*/

inline_static void
reset_Checker (Checker * checker)
{
  release_FastStack (checker->compsat, &checker->assignment);
  reset_FastStack (checker->compsat, &checker->stack, 0);
}

/*------------------------------------------------------------------------*/

inline_static void
release_Checker (Checker * checker)
{
  release_FastStack (checker->compsat, &checker->units);
  release_FastStack (checker->compsat, &checker->assignment);
  release_FastStack (checker->compsat, &checker->stack);
  release_stack2 (checker->compsat, checker->occurence, checker->compsat->size);
}

/*------------------------------------------------------------------------*/

inline_static FastStack *
occurence (Checker * checker, int lit)
{
  return checker->occurence[SGN (lit)] + ABS (lit);
}

/*------------------------------------------------------------------------*/

inline_static void
add_Checker (Checker * checker, int clause_idx)
{
  CursorCompactStack (&checker->compsat->literals, it, clause_idx - 1, 1);
  FastStack *stack;
  int i, lit;

  lit = next_CompactStack (it);
  assert (!lit);
  i = clause_idx;

  while ((lit = next_CompactStack (it)))
    {
      stack = occurence (checker, lit);
      push_FastStack (checker->compsat, stack, i++);
    }
}

/*------------------------------------------------------------------------*/

inline_static int
deref_Checker (Checker * checker, int lit)
{
  int res;

  res = get_FastStack (&checker->assignment, ABS (lit));

  if (!res)
    return lit;

  if (SGN (lit))
    res = -res;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
assign_Checker (Checker * checker, int lit)
{
  int val;

  val = SGN (lit) ? FALSE : TRUE;
  set_FastStack (checker->compsat, &checker->assignment, ABS (lit), val);
}

/*------------------------------------------------------------------------*/

inline_static void
check_is_implied (Checker * checker)
{
  int occidx, conflict, lit, found, assignment, tmp, dir, steps;
  CompSat *compsat;
  FastStack *occ;

  compsat = checker->compsat;

  {
    IteratorFastStack (&checker->units, unitit, 1);
    while (!done_FastStack (unitit))
      push_FastStack (compsat, &checker->stack, next_FastStack (unitit));
  }

  {
    IteratorFastStack (&compsat->clause, clauseit, 1);
    while (!done_FastStack (clauseit))
      push_FastStack (compsat, &checker->stack, -next_FastStack (clauseit));
  }

  conflict = 0;
  steps = 0;

  while (!conflict && !is_empty_FastStack (&checker->stack))
    {
      steps++;

      lit = pop_FastStack (compsat, &checker->stack);
      tmp = deref_Checker (checker, lit);

      if ((conflict = (tmp == FALSE)))
	break;

      if (tmp == TRUE)
	continue;

      assign_Checker (checker, lit);
      occ = occurence (checker, -lit);

      {
	IteratorFastStack (occ, occit, 1);
	while (!conflict && !done_FastStack (occit))
	  {
	    occidx = next_FastStack (occit);
	    assignment = 0;
	    found = 0;

	    for (dir = -1; dir <= 1; dir += 2)
	      {
		/* Be sure to visit all literals since 'compsat' may reorder
		 * literals and 'occidx' may not point to '-lit' anymore.
		 */
		CursorCompactStack (&compsat->literals, litit, occidx, dir);
		while (found < 2 && (lit = next_CompactStack (litit)))
		  {
		    tmp = deref_Checker (checker, lit);
		    if (tmp == TRUE)
		      found = 2;
		    else if (tmp != FALSE)
		      {
			assert (lit == tmp);
			assignment = tmp;
			found++;
		      }
		  }
		occidx++;
	      }

	    if (!found)
	      conflict = 1;
	    else if (found == 1)
	      push_FastStack (compsat, &checker->stack, assignment);
	  }
      }
    }

#ifdef COMPSAT_LOG
  log (compsat, 3, "%d CHECK %d\n", compsat->level, steps);
#endif
  assert (conflict);
  reset_Checker (checker);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
connect_literal (CompSat * compsat, int lit_idx)
{
  int lit;

#ifdef COMPSAT_LOG
  log_assignment (compsat, 6, "CONNECT", lit_idx);
#endif

  lit = get_CompactStack (&compsat->literals, lit_idx);
  push_CompactStack (compsat, get_watched (compsat, lit), lit_idx);
}

/*------------------------------------------------------------------------*/
#if defined(COMPSAT_LOG) || defined(COMPSAT_CHECK)
/*------------------------------------------------------------------------*/

void
COMPSAT_dump_clause (CompSat * compsat, FILE * file, int clause_idx)
{
  CompactStack *literals = &compsat->literals;
  int i, j;

  /* The following probably is a bad style of defensive programming.
   * However it makes it easy to call 'COMPSAT_dump_clause' from a debugger,
   * since 'stdout' may not be accessible as a symbolic name.
   */
  if (!file)
    file = stdout;

  while (get_CompactStack (literals, clause_idx - 1))
    clause_idx--;

  for (i = 0; get_CompactStack (literals, clause_idx + i); i++)
    ;

  fprintf (file, "CLAUSE [%d,%d]", clause_idx, clause_idx + i - 1);

  for (j = clause_idx; j < clause_idx + i; j++)
    fprintf (file, " %d", get_CompactStack (literals, j));

  fprintf (file, "\n");
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_LOG
/*------------------------------------------------------------------------*/

inline_static void
log_clause (CompSat * compsat, int level, int clause_idx)
{
  FILE *file;

  if (compsat->verbosity >= level)
    {
      file = log_file (compsat);
      fprintf (file, COMPSAT_PREFIX "%d ", compsat->level);
      COMPSAT_dump_clause (compsat, file, clause_idx);
    }
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
count_literals (CompSat * compsat, int clause_idx)
{
  CursorCompactStack (&compsat->literals, it, clause_idx, 1);
  int lit, i;

  for (i = 0; (lit = next_CompactStack (it)); i++)
    inc_ref (compsat, lit);

  compsat->stats.clauses++;
  compsat->stats.literals += i;
}

/*------------------------------------------------------------------------*/

inline_static void
connect_clause (CompSat * compsat, int clause_idx)
{
  int i, first_lit;

  if (!(first_lit = get_CompactStack (&compsat->literals, clause_idx)))
    {
#ifdef COMPSAT_LOG
      log (compsat, 3, "%d EMPTY\n", compsat->level);
#endif
      compsat->contains_empty_clause = 1;
      reset_CompactStack (compsat, &compsat->literals, clause_idx);
    }
  else if (!get_CompactStack (&compsat->literals, clause_idx + 1))
    {
#ifdef COMPSAT_LOG
      log (compsat, 3, "%d UNIT %d\n", compsat->level, first_lit);
#endif
      push_FastStack (compsat, &compsat->units, first_lit);
      reset_CompactStack (compsat, &compsat->literals, clause_idx);
#ifdef COMPSAT_CHECK
      if (compsat->checker.enabled)
	push_FastStack (compsat, &compsat->checker.units, first_lit);
#endif
    }
  else
    {
      CursorCompactStack (&compsat->literals, it, clause_idx, 1);
#ifdef COMPSAT_LOG
      log_clause (compsat, 3, clause_idx);
#endif

      for (i = 0; i < MAX_WATCHED && next_CompactStack (it); i++)
	connect_literal (compsat, i + clause_idx);

#ifdef COMPSAT_CHECK
      if (compsat->checker.enabled)
	{
	  check_clause (compsat, clause_idx);
	  add_Checker (&compsat->checker, clause_idx);
	}
#endif
    }
}

/*------------------------------------------------------------------------*/

inline_static int
add (CompSat * compsat)
{
  int clause_idx;

  push_FastStack (compsat, &compsat->clause, 0);
  clause_idx = copy_clause (compsat, compsat->clause.start);
  order_clause (compsat, clause_idx);
  count_literals (compsat, clause_idx);
  connect_clause (compsat, clause_idx);
  reset_FastStack (compsat, &compsat->clause, 0);

  return clause_idx;
}

/*------------------------------------------------------------------------*/

inline_static int
deref (CompSat * compsat, int lit)
{
  int res;

  res = get_assignment (compsat, ABS (lit));

  if (!res)
    return lit;

  if (SGN (lit))
    res = -res;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
is_assigned (CompSat * compsat, int lit)
{
  return get_assignment (compsat, ABS (lit)) != 0;
}

/*------------------------------------------------------------------------*/

inline_static int
propagate (CompSat * compsat, int lit_idx, int *remove_ptr)
{
  CursorCompactStack (&compsat->literals, it, lit_idx - 1, 1);
  int lit, tmp, assignment, old_lit;
  void *old, *last;

  assert (MAX_WATCHED == 2);
  *remove_ptr = 0;

  lit = next_CompactStack (it);
  save_CursorCompactStack (it, old);
  old_lit = next_CompactStack (it);
  assert (deref (compsat, old_lit) == FALSE);

#ifdef COMPSAT_LOG
  log (compsat, 6, "%d PROPAGATE %d %d\n", compsat->level, old_lit, lit_idx);
#endif

  if (lit)
    {
      assert (!get_CompactStack (&compsat->literals, lit_idx - 2));
      tmp = deref (compsat, lit);
      if (tmp == TRUE)
	return 0;
      if (tmp == FALSE)
	return lit_idx;
      assignment = lit_idx - 1;
    }
  else if ((lit = next_CompactStack (it)))
    {
      tmp = deref (compsat, lit);
      if (tmp == TRUE)
	return 0;
      if (tmp == FALSE)
	return lit_idx;
      assignment = lit_idx + 1;
    }
  else
    return lit_idx;

  /* Now search through the non watched literals of the clause for a non
   * FALSE assigned literal.  If one is found, then disconnect the 'old'
   * literal and connect the new literal.  In addition swap the position of
   * the non false literal with the start literal.
   */
  save_CursorCompactStack (it, last);
  while ((lit = next_CompactStack (it)))
    {
      tmp = deref (compsat, lit);

      if (tmp != FALSE)
	{
#ifdef COMPSAT_LOG
	  log_assignment (compsat, 6, "DISCONNECT", lit_idx);
#endif
	  swap_CompactStack (&compsat->literals, it, old, last);
	  *remove_ptr = 1;
	  connect_literal (compsat, lit_idx);

	  return 0;
	}

      save_CursorCompactStack (it, last);
    }

  compsat->stats.assignments++;
  enqueue (compsat, &compsat->queue, assignment);

  return 0;
}

/*------------------------------------------------------------------------*/

inline_static void
assign (CompSat * compsat, int assignment)
{
  int idx, sign, lit, reason;

#ifdef COMPSAT_LOG
  log_assignment (compsat, 4, "ASSIGN", assignment);
#endif
  split_assignment (compsat, assignment, &lit, &reason);
  idx = ABS (lit);
  sign = SGN (lit);

  push_FastStack (compsat, &compsat->trail, lit);

  set_assignment (compsat, idx, sign ? FALSE : TRUE);
  set_reason (compsat, idx, reason);
  set_level (compsat, idx, compsat->level);
}

/*------------------------------------------------------------------------*/

inline_static int
deduce (CompSat * compsat)
{
  int assignment, tmp, lit, reason, conflict, remove, *p, *last;
  CompactStack *stack;

  assignment = dequeue (compsat, &compsat->queue);
  split_assignment (compsat, assignment, &lit, &reason);
  tmp = deref (compsat, lit);
  conflict = 0;

  if (tmp == FALSE)
    conflict = assignment;
  else if (tmp != TRUE)
    {
      assign (compsat, assignment);

      stack = get_watched (compsat, -lit);
      p = stack->start;
      last = stack->top;
      conflict = 0;

      while (!conflict && p < last)
	{
	  conflict = propagate (compsat, *p, &remove);

	  if (remove)
	    *p = *--last;
	  else
	    p++;
	}

      reset_FastStack (compsat, stack, last - stack->start);
    }

#ifdef COMPSAT_LOG
  if (conflict)
    {
      log (compsat, 3, "%d FALSE %s %d\n", compsat->level,
	   (conflict < 0) ? "LITERAL" : "CLAUSE",
	   (conflict < 0) ? lit : conflict);
    }
#endif

  if (conflict)
    compsat->stats.conflicts++;

  return conflict;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/

inline_static void
check_is_satisfied (CompSat * compsat)
{
  IteratorCompactStack (&compsat->literals, it, 1);
  int clause_satisfied, lit, tmp;

  tmp = next_CompactStack (it);
  assert (!tmp);

  while (!done_CompactStack (it))
    {
      clause_satisfied = 0;

      while ((lit = next_CompactStack (it)))
	{
	  if (!clause_satisfied)
	    {
	      tmp = deref (compsat, lit);
	      clause_satisfied = (tmp == TRUE);
	    }
	}

      assert (clause_satisfied);
    }
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
report (CompSat * compsat)
{
  double mb;

  if (compsat->verbosity >= 1)
    {
      compsat->stats.reported++;

      if(compsat->stats.reported == 1)
        {
	  log (compsat, 1, "\n");
	  log (compsat, 1, "   seconds         decisions              MB\n");
	  log (compsat, 1, "          conflicts         assignments\n");
	  log (compsat, 1, "\n");
	}

      update_time (compsat);
      mb = COMPSAT_bytes (compsat);
      mb /= 1024.0;
      mb /= 1024.0;

      log (compsat, 1,
	   "%8.2f %8.0f %8.0f %10.0f %7.1f\n",
	   compsat->stats.time,
	   (double) compsat->stats.conflicts,
	   (double) compsat->stats.decisions,
	   (double) compsat->stats.assignments,
	   mb);
    }
}

/*------------------------------------------------------------------------*/

inline_static int
decide (CompSat * compsat)
{
  int *p, *order, pos, *end, lit, assignment;

  if (done_Counter (&compsat->counters.report))
    report (compsat);

  compsat->stats.decisions++;
  order = compsat->order.start;
  end = compsat->order.top;

  for (p = order + compsat->unassigned; p < end; p++)
    {
      lit = *p;
      if (!is_assigned (compsat, lit))
	{
	  compsat->level++;
	  push_FastStack (compsat, &compsat->control,
			  count_FastStack (&compsat->trail));
	  assert (compsat->level == count_FastStack (&compsat->control));
	  pos = p - order;
#ifdef COMPSAT_LOG
	  log (compsat, 3, "%d DECIDE %d %d\n", compsat->level, lit, pos);
#endif
	  assignment = unit2assignment (compsat, lit);
	  enqueue (compsat, &compsat->queue, assignment);
	  compsat->unassigned = pos + 1;
	  return 1;
	}
    }

  compsat->unassigned = count_FastStack (&compsat->order);
#ifdef COMPSAT_LOG
  log (compsat, 3, "%d ALLASSIGNED\n", compsat->level);
#endif
#ifdef COMPSAT_CHECK
  if (compsat->checker.enabled)
    check_is_satisfied (compsat);
#endif

  return 0;
}

/*------------------------------------------------------------------------*/

inline_static void
units2queue (CompSat * compsat)
{
  IteratorPointerStack (&compsat->units, it, 1);
  int lit;

  while (!done_PointerStack (it))
    {
      lit = next_PointerStack (it);
      enqueue (compsat, &compsat->queue, unit2assignment (compsat, lit));
    }
}

/*------------------------------------------------------------------------*/

inline_static int
cmp (CompSat * compsat, int a, int b)
{
  int score[2], idx[2], sign[2];

  score[0] = get_score (compsat, a);
  score[1] = get_score (compsat, b);

  if (score[0] < score[1])
    return -1;
  if (score[0] > score[1])
    return 1;

  idx[0] = ABS (a);
  idx[1] = ABS (b);

  if (idx[0] > idx[1])
    return -1;
  if (idx[0] < idx[1])
    return 1;

  sign[0] = SGN (a);
  sign[1] = SGN (b);

  if (sign[0] < sign[1])
    return -1;
  if (sign[0] > sign[1])
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

inline_static void
swap (int *a, int *b)
{
  int tmp;

  tmp = *a;
  *a = *b;
  *b = tmp;
}

/*------------------------------------------------------------------------*/

inline_static void
cmpswap (CompSat * compsat, int *a, int *b)
{
  if (cmp (compsat, *a, *b) < 0)
    swap (a, b);
}

/*------------------------------------------------------------------------*/

inline_static int
partition (CompSat * compsat, int *a, int l, int r)
{
  int lit, i, j;

  i = l - 1;
  j = r;
  lit = a[r];

  for (;;)
    {
      while (cmp (compsat, a[++i], lit) > 0)
	;

      while (cmp (compsat, lit, a[--j]) > 0)
	if (j == l)
	  break;

      if (i >= j)
	break;

      swap (a + i, a + j);
    }

  swap (a + i, a + r);

  return i;
}

/*------------------------------------------------------------------------*/

inline_static void
insertion (CompSat * compsat, int *a, int l, int r)
{
  int lit, i, j;

  for (i = r; i > l; i--)
    cmpswap (compsat, a + i - 1, a + i);

  for (i = l + 2; i <= r; i++)
    {
      j = i;
      lit = a[i];
      while (cmp (compsat, lit, a[j - 1]) > 0)
	{
	  a[j] = a[j - 1];
	  j--;
	}
      a[j] = lit;
    }
}

/*------------------------------------------------------------------------*/

inline_static void
quicksort (CompSat * compsat, int *a, int l, int r)
{
  int i, m, ll, rr;
  FastStack *stack;

  if (r - l <= ISORT_LIMIT)
    return;

  stack = &compsat->stack;
  reset_FastStack (compsat, stack, 0);

  for (;;)
    {
      m = (l + r) / 2;
      swap (a + m, a + r - 1);
      cmpswap (compsat, a + l, a + r - 1);
      cmpswap (compsat, a + l, a + r);
      cmpswap (compsat, a + r - 1, a + r);
      i = partition (compsat, a, l + 1, r - 1);

      if (i - l < r - i)
	{
	  ll = i + 1;
	  rr = r;
	  r = i - 1;
	}
      else
	{
	  ll = l;
	  rr = i - 1;
	  l = i + 1;
	}

      if (r - l > ISORT_LIMIT)
	{
	  assert (rr - ll > ISORT_LIMIT);
	  push_FastStack (compsat, stack, ll);
	  push_FastStack (compsat, stack, rr);
	}
      else if (rr - ll > ISORT_LIMIT)
	{
	  l = ll;
	  r = rr;
	}
      else if (count_FastStack (stack))
	{
	  r = pop_FastStack (compsat, stack);
	  l = pop_FastStack (compsat, stack);
	}
      else
	break;
    }
}

/*------------------------------------------------------------------------*/

inline_static void
hypersort (CompSat * compsat, int *a, int n)
{
  quicksort (compsat, a, 0, n - 1);
  insertion (compsat, a, 0, n - 1);
}

/*------------------------------------------------------------------------*/

inline_static void
adjust (CompSat * compsat)
{
  int i, lit, last, last_non_zero_pos, first_unassigned;

  last = compsat->zero - 1;
  last_non_zero_pos = -1;
  first_unassigned = -1;

  for (i = 0; i < compsat->zero; i++)
    {
      lit = get_FastStack (&compsat->order, i);
      if (first_unassigned < 0 && !is_assigned (compsat, lit))
	first_unassigned = i;
      if (get_score (compsat, lit))
	last_non_zero_pos = i;
      set_pos (compsat, lit, i);
    }

  compsat->zero = last_non_zero_pos + 1;
  compsat->unassigned = (first_unassigned < 0) ? 0 : first_unassigned;
}

/*------------------------------------------------------------------------*/

inline_static void
sort (CompSat * compsat)
{
#ifdef COMPSAT_LOG
  log (compsat, 2, "%d SORT %d\n", compsat->level, compsat->zero);
#endif
  hypersort (compsat, compsat->order.start, compsat->zero);
  adjust (compsat);
}

/*------------------------------------------------------------------------*/

inline_static void
order (CompSat * compsat)
{
  int idx, lit, pos, ref, sign;

  pos = 0;
  for (idx = 1; idx <= compsat->max; idx++)
    for (sign = 1; sign >= 0; sign--)
      {
	lit = (sign ? -1 : 1) * idx;
	set_FastStack (compsat, &compsat->order, pos, lit);
	set_pos (compsat, lit, pos);
	ref = get_ref (compsat, lit);
	set_score (compsat, lit, ref);
	pos++;
      }
  assert (pos == count_FastStack (&compsat->order));
  compsat->unassigned = 0;
  compsat->zero = pos;

  sort (compsat);
}

/*------------------------------------------------------------------------*/

inline_static void
rescore (CompSat * compsat)
{
  int pos, lit, score, zero;

#ifdef COMPSAT_LOG
  log (compsat, 2, "%d RESCORE %.0f\n",
       compsat->level, compsat->counters.rescore.finished);
#endif

  zero = -1;
  for (pos = compsat->zero - 1; pos >= 0; pos--)
    {
      lit = get_FastStack (&compsat->order, pos);
      score = get_score (compsat, lit);
      score /= AGING_DIVISOR;
      if (score && pos > zero)
	zero = pos + 1;
      set_score (compsat, lit, score);
    }

  compsat->zero = zero;
  sort (compsat);
}

/*------------------------------------------------------------------------*/

inline_static int
expand (CompSat * compsat, int reason, int paths)
{
  int lit, new_lit, idx, dir, level, tmp;

  for (dir = -1; dir <= 1; dir += 2)
    {
      CursorCompactStack (&compsat->literals, it, reason + (dir == 1), dir);
      while ((lit = next_CompactStack (it)))
	{
	  idx = ABS (lit);
	  level = get_level (compsat, idx);

	  if (level > 0 && !get_mark (compsat, idx))
	    {
	      set_mark (compsat, idx);
	      push_FastStack (compsat, &compsat->visited, idx);

	      tmp = deref (compsat, lit);
	      add_score (compsat, (tmp == TRUE) ? -lit : lit, AGING_DIVISOR);

	      if (level == compsat->level)
		paths++;
	      else
		{
		  new_lit = (tmp == TRUE) ? -lit : lit;
		  push_FastStack (compsat, &compsat->clause, new_lit);
		}
	    }
	}
    }

  return paths;
}

/*------------------------------------------------------------------------*/

inline_static void
analyze (CompSat * compsat, int conflict)
{
  int paths, *p, lit, idx, reason;

  paths = expand (compsat, conflict, 0);
  p = compsat->trail.top;

  do
    {
      lit = *--p;
      idx = ABS (lit);

      if (get_mark (compsat, idx))
	{
	  if (--paths)
	    {
	      reason = get_reason (compsat, idx);
	      paths = expand (compsat, reason, paths);
	    }
	  else
	    push_FastStack (compsat, &compsat->clause, -lit);
	}
    }
  while (paths);

  unmark_visited (compsat);

#ifdef COMPSAT_LOG
  if (compsat->level >= 2 && get_reason (compsat, idx))
    log (compsat, 3, "%d UIP %d\n", compsat->level, lit);
#endif
#ifdef COMPSAT_CHECK
  if (compsat->checker.enabled)
    check_is_implied (&compsat->checker);
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
drive (CompSat * compsat, int clause_idx)
{
  CursorCompactStack (&compsat->literals, it, clause_idx, 1);
  int lit, next_lit, res, count, level;

  lit = next_CompactStack (it);
  assert (lit);
  next_lit = next_CompactStack (it);

  if (next_lit)
    {
      assert (clause_idx < count_CompactStack (&compsat->literals));

      for (count = clause_idx, res = 0; !res && lit; count++)
	{
	  level = get_level (compsat, ABS (lit));
	  if (level == compsat->level)
	    res = count;
	  else if ((lit = next_lit))
	    next_lit = next_CompactStack (it);
	}
    }
  else
    {
      assert (clause_idx == count_CompactStack (&compsat->literals));
      res = unit2assignment (compsat, lit);
    }

  assert (res);

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
jump (CompSat * compsat, int clause_idx)
{
  CursorCompactStack (&compsat->literals, it, clause_idx, 1);
  int tmp, res, lit;

  res = 0;

  while ((lit = next_CompactStack (it)))
    {
      tmp = get_level (compsat, ABS (lit));
      if (tmp < compsat->level && tmp > res)
	res = tmp;
    }

#ifdef COMPSAT_LOG
  log (compsat, 3, "%d BACK%s %d\n",
       compsat->level, (res < compsat->level - 1) ? "JUMP" : "TRACK", res);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
untrail (CompSat * compsat, int level)
{
  int *start, *last, *p, frame_start, pos_in_order, lit;

  frame_start = get_FastStack (&compsat->control, level);
  start = compsat->trail.start + frame_start;
  last = compsat->trail.top - 1;

  for (p = last; p >= start; p--)
    {
      lit = *p;
#ifdef COMPSAT_LOG
      log (compsat, 4, "%d UNASSIGN %d\n", compsat->level, lit);
#endif
      set_assignment (compsat, ABS (lit), 0);
      pos_in_order = get_pos (compsat, lit);
      if (pos_in_order < compsat->unassigned)
	compsat->unassigned = pos_in_order;
    }

  compsat->level = level;
  reset_FastStack (compsat, &compsat->control, level);
  reset_FastStack (compsat, &compsat->trail, frame_start);
}

/*------------------------------------------------------------------------*/

inline_static void
init_sat (CompSat * compsat)
{
  assert (is_empty_Queue (&compsat->queue));
  assert (!compsat->contains_empty_clause);
  units2queue (compsat);
  order (compsat);
}

/*------------------------------------------------------------------------*/

static void
check_all_unassigned (CompSat * compsat)
{
#ifdef COMPSAT_CHECK
  int i, tmp;
  
  for (i = 1; i <= compsat->max; i++)
    {
      if (get_level (compsat, i) > 0)
      {
      tmp = get_assignment (compsat, i);
      assert (tmp == 0);
      }
    }
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
restart (CompSat * compsat)
{
  assert (compsat->restart_at_next_decision);
  compsat->restart_at_next_decision = 0;
#ifdef COMPSAT_LOG
  log (compsat, 2, "%d RESTART %.0f\n",
       compsat->level, compsat->counters.restart.finished);
#endif
  untrail (compsat, 0);
  check_all_unassigned (compsat);
  init_sat (compsat);
}

/*------------------------------------------------------------------------*/

inline_static int
backtrack (CompSat * compsat, int conflict)
{
  int clause, assignment, level;

  reset_Queue (compsat, &compsat->queue);

  if (!compsat->level)
    return 0;

  analyze (compsat, conflict);
  clause = add (compsat);

  assignment = drive (compsat, clause);
  enqueue (compsat, &compsat->queue, assignment);

  level = jump (compsat, clause);
  untrail (compsat, level);

  if (done_Counter (&compsat->counters.rescore))
    rescore (compsat);

  if (!compsat->restart_at_next_decision &&
      done_Counter (&compsat->counters.restart))
    compsat->restart_at_next_decision = 1;

  return 1;
}

/*------------------------------------------------------------------------*/

inline_static int
reached_decision_limit (CompSat * compsat)
{
  if (compsat->limits.decisions < 0)
    return 0;
  else
    return compsat->stats.decisions > compsat->limits.decisions;
}

/*------------------------------------------------------------------------*/

inline_static int
reached_time_limit (CompSat * compsat)
{
  double limit;
  int res;

  limit = compsat->limits.time;
  if (limit < 0)
    return 0;

  update_time (compsat);
  res = (compsat->stats.time >= limit);

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
sat (CompSat * compsat)
{
  int conflict;

  if (compsat->contains_empty_clause)
    return COMPSAT_UNSATISFIABLE;

  init_sat (compsat);

  for (;;)
    {
      if (compsat->restart_at_next_decision &&
	  is_empty_Queue (&compsat->queue))
	restart (compsat);

      if (is_empty_Queue (&compsat->queue))
	{
	  if (!decide (compsat))
	    return COMPSAT_SATISFIABLE;

	  if (reached_decision_limit (compsat))
	    return COMPSAT_DECISION_OUT;

	  if (reached_time_limit (compsat))
	    return COMPSAT_TIME_OUT;
	}

      conflict = deduce (compsat);

      if (conflict && !backtrack (compsat, conflict))
	return COMPSAT_UNSATISFIABLE;
    }
}

/*------------------------------------------------------------------------*/

inline_static void
enter (CompSat * compsat)
{
  if (!compsat->internal++)
    compsat->entered = COMPSAT_get_time ();
}

/*------------------------------------------------------------------------*/

inline_static void
leave (CompSat * compsat)
{
  assert (compsat->internal > 0);

  if (compsat->internal == 1)
    {
      compsat->stats.time += delta_time (compsat);
      compsat->entered = 0;
    }

  compsat->internal--;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

inline_static int
err_aux (CompSat * compsat, const char *fmt, va_list ap)
{
  FILE *file;

  file = log_file (compsat);
  fprintf (file, "*** compsat: ");
  vfprintf (file, fmt, ap);
  fputc ('\n', file);

  return COMPSAT_ABORT;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

inline_static int
err (CompSat * compsat, const char *fmt, ...)
{
  va_list ap;
  int res;

  va_start (ap, fmt);
  res = err_aux (compsat, fmt, ap);
  va_end (ap);

  return res;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

CompSat *
COMPSAT_new (void)
{
  CompSat *res;

  res = (CompSat *) calloc (1, sizeof (*res));
  push_CompactStack (res, &res->literals, 0);

  res->stats.time = 0;
  res->stats.decisions = 0;
  res->stats.conflicts = 0;
  res->stats.assignments = 0;

  res->limits.time = -1;
  res->limits.decisions = -1;
  res->limits.bytes = -1;

  init_Counter (&res->counters.rescore, 128 * AGING_DIVISOR, 1, 0);
  init_Counter (&res->counters.restart, 10000, 1.1, 1);
  init_Counter (&res->counters.report, 1000, 1.4142, 1);

  res->verbosity=1;

#if 0
  init_Counter (&res->counters.rescore, 100, 1, 0);
  init_Counter (&res->counters.restart, 200, 1.1, 1);
#endif

#ifdef COMPSAT_CHECK
  init_Checker (&res->checker, res);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

void
COMPSAT_set_leak_hook (CompSat * compsat, size_t * leaked_ptr)
{
  *leaked_ptr = 0;;
#ifdef COMPSAT_CHECK
  compsat->leaked_ptr = leaked_ptr;
#endif
}

/*------------------------------------------------------------------------*/

void
COMPSAT_delete (CompSat * compsat)
{
  if (!compsat)
    return;

#ifdef COMPSAT_CHECK
  release_Checker (&compsat->checker);
#endif

  if (compsat->log)
    fclose (compsat->log);

  if (compsat->out)
    fclose (compsat->out);

  release_CompactStack (compsat, &compsat->literals);
  release_FastStack (compsat, &compsat->units);
  release_Queue (compsat, &compsat->queue);
  release_FastStack (compsat, &compsat->trail);
  release_FastStack (compsat, &compsat->control);
  release_FastStack (compsat, &compsat->order);
  release_FastStack (compsat, &compsat->clause);
  release_FastStack (compsat, &compsat->visited);
  release_FastStack (compsat, &compsat->stack);
  delete_string (compsat, compsat->error);
  delete_string (compsat, compsat->name);

#ifdef COMPSAT_SLICE_VARIABLE
  release_stack2 (compsat, compsat->watched, compsat->size);
  release_FastStack (compsat, &compsat->assignment);
  release_FastStack (compsat, &compsat->levels);
  release_FastStack (compsat, &compsat->reason);
  release_FastStack (compsat, &compsat->score[0]);
  release_FastStack (compsat, &compsat->score[1]);
  release_FastStack (compsat, &compsat->ref[0]);
  release_FastStack (compsat, &compsat->ref[1]);
  release_FastStack (compsat, &compsat->pos[0]);
  release_FastStack (compsat, &compsat->pos[1]);
#else
  release_variables (compsat);
#endif

  release_FastStack (compsat, &compsat->mark[0]);
  release_FastStack (compsat, &compsat->mark[1]);
  release_FastStack (compsat, &compsat->used);

#ifdef COMPSAT_CHECK
  if (compsat->leaked_ptr)
    *compsat->leaked_ptr = compsat->stats.bytes;
#endif

  free (compsat);
}

/*------------------------------------------------------------------------*/

inline_static int
count_clauses (CompSat * compsat)
{
  IteratorCompactStack (&compsat->literals, it, 1);
  int lit, res;

  res = 0;
  lit = next_CompactStack (it);
  assert (!lit);

  while (!done_CompactStack (it))
    {
      lit = next_CompactStack (it);
      if (!lit)
	res++;
    }

  return res;
}

/*------------------------------------------------------------------------*/

void
COMPSAT_print (CompSat * compsat, FILE * file)
{
  int count, lit;

  enter (compsat);

  if (compsat->contains_empty_clause)
    fprintf (file, "p cnf 0 1\n0\n");
  else
    {
      count = count_clauses (compsat);
      count += count_FastStack (&compsat->units);

      fprintf (file, "p cnf %d %d\n", compsat->max, count);

      {
	IteratorCompactStack (&compsat->literals, litit, 1);
	lit = next_CompactStack (litit);
	assert (!lit);
	while (!done_CompactStack (litit))
	  {
	    lit = next_CompactStack (litit);
	    fprintf (file, "%d%c", lit, lit ? ' ' : '\n');
	  }
      }

      {
	IteratorFastStack (&compsat->units, unitit, 1);
	while (!done_FastStack (unitit))
	  fprintf (file, "%d 0\n", next_FastStack (unitit));
      }
    }

  leave (compsat);
}

/*------------------------------------------------------------------------*/

inline_static void
COMPSAT_add (CompSat * compsat, const int *clause)
{
  int clause_idx;

  enter (compsat);

  insert_variables (compsat, clause);
  clause_idx = simplify_clause (compsat, clause);

  if (clause_idx)
    {
      order_clause (compsat, clause_idx);
      count_literals (compsat, clause_idx);
      connect_clause (compsat, clause_idx);
    }

  leave (compsat);
}

/*------------------------------------------------------------------------*/

const char *
COMPSAT_read (CompSat * compsat, const char *name)
{
  int added_literals, added_clauses, added_variables;
  int old_literals, old_variables;
  double time_before_read, delta;
  const char *error;
  const int *clause;
  Dimacs *parser;
  DimacsMem mem;

  enter (compsat);

  memset (&mem, 0, sizeof (mem));
  mem.mgr = compsat;
  mem.new = (DimacsNew) new;
  mem.delete = (DimacsDelete) delete;

  log (compsat, 1, "reading %s\n", name);
  time_before_read = COMPSAT_get_time ();
  old_literals = compsat->stats.literals;
  old_variables = compsat->stats.variables;

  /* Since trivial clauses are not really added we have to count the number
   * of clauses found in the file.
   */
  added_clauses = 0;
  parser = DIMACS_new (name, &mem);
  while ((clause = DIMACS_read (parser)))
    {
      COMPSAT_add (compsat, clause);
      added_clauses++;
    }

  /* Variables of trivial clauses are still added.    For literals this is
   * not the case but we just don't care about literals in trivial clauses.
   */
  added_literals = compsat->stats.literals - old_literals;
  added_variables = compsat->stats.variables - old_variables;

  error = DIMACS_error (parser);
  if (!error)
    {
      if (DIMACS_specified_num_clauses (parser) < added_clauses)
	error = "more clauses found than specified";

      if (DIMACS_specified_num_clauses (parser) > added_clauses)
	error = "less clauses found than specified";

      /* TODO: if we want to allow multiple read we have to fix this test.
       */
      if (DIMACS_specified_max_variable (parser) < compsat->max)
	error = "specified maximal variable index exceeded";
    }

  if (error)
    {
      delete_string (compsat, compsat->error);
      compsat->error = copy_string (compsat, error);
    }
  else
    {
      delta = COMPSAT_get_time() - time_before_read;
      delta = (delta < 0) ? 0 : delta;

      log (compsat, 1, "read %d literals in %0.2f seconds\n",
	   added_literals, delta);

      log (compsat, 1, "found %d new variables in %d clauses\n",
	   added_variables, added_clauses);
    }

#ifdef COMPSAT_CHECK
  if (!error)
    check (compsat);
#endif

  DIMACS_delete (parser);
  leave (compsat);

  return compsat->error;
}

/*------------------------------------------------------------------------*/

size_t
COMPSAT_bytes (CompSat * compsat)
{
  return compsat->stats.max_bytes + sizeof (*compsat);
}

/*------------------------------------------------------------------------*/

double
COMPSAT_seconds (CompSat * compsat)
{
  if (compsat->internal)
    update_time (compsat);

  return compsat->stats.time;
}

/*------------------------------------------------------------------------*/

void
COMPSAT_adjust_timer (CompSat * compsat)
{
  update_time (compsat);
}

/*------------------------------------------------------------------------*/

void
COMPSAT_print_seconds (FILE * file, double seconds)
{
  fprintf (file, "%.2f sec", seconds);
}

/*------------------------------------------------------------------------*/

void
COMPSAT_print_bytes (FILE * file, size_t bytes)
{
  if (bytes < (1 << 10))
    fprintf (file, "%.0f bytes", (double) bytes);
  else if (bytes < (1 << 20))
    fprintf (file, "%.1f KB", ((double) bytes) / (double) (1 << 10));
  else if (bytes < (1 << 30))
    fprintf (file, "%.1f MB", ((double) bytes) / (double) (1 << 20));
  else
    fprintf (file, "%.1f GB", ((double) bytes) / (double) (1 << 30));
}

/*------------------------------------------------------------------------*/

FILE *
COMPSAT_out_file (CompSat * compsat)
{
  return out_file (compsat);
}

/*------------------------------------------------------------------------*/

FILE *
COMPSAT_log_file (CompSat * compsat)
{
  return log_file (compsat);
}

/*------------------------------------------------------------------------*/

FILE *
COMPSAT_err_file (CompSat * compsat)
{
  return log_file (compsat);	/* no explicit error file */
}

/*------------------------------------------------------------------------*/

void
COMPSAT_stats (CompSat * compsat, FILE * file)
{
  enter (compsat);

  fprintf (file,
	   COMPSAT_PREFIX
	   "%.0f conflicts, %.0f decisions, %.0f assignments\n",
	   (double) compsat->stats.conflicts,
	   (double) compsat->stats.decisions,
	   (double) compsat->stats.assignments);

  fprintf (file,
	   COMPSAT_PREFIX
	   "%d variables, %d clauses, %d literals\n",
	   compsat->stats.variables,
	   compsat->stats.clauses, compsat->stats.literals);

  fputs (COMPSAT_PREFIX, file);
  COMPSAT_print_seconds (file, COMPSAT_seconds (compsat));
  fputs (", ", file);
  COMPSAT_print_bytes (file, COMPSAT_bytes (compsat));
  fputc ('\n', file);

  leave (compsat);
}

/*------------------------------------------------------------------------*/

const char *
COMPSAT_res2str (int res)
{
  switch (res)
    {
    case COMPSAT_SATISFIABLE:
      return ("SATISFIABLE");
      break;
    case COMPSAT_UNSATISFIABLE:
      return ("UNSATISFIABLE");
      break;
    case COMPSAT_TIME_OUT:
      return ("TIMEOUT");
      break;
    case COMPSAT_SPACE_OUT:
      return ("SPACEOUT");
      break;
    case COMPSAT_DECISION_OUT:
      return ("DECISIONOUT");
      break;
    case COMPSAT_EXIT:
      return ("EXIT");
      break;
    default:
      assert (res == COMPSAT_ABORT);
      return ("ABORT");
      break;
    }
}

/*------------------------------------------------------------------------*/

int
COMPSAT_sat (CompSat * compsat)
{
  int res;

  enter (compsat);
  res = sat (compsat);
  if(compsat->stats.reported > 0)
    {
      report (compsat);
      log (compsat, 0, "\n");
    }
  leave (compsat);

  return res;
}

/*------------------------------------------------------------------------*/

#define LINE_WIDTH 80
#define MAX_POS 77

void
COMPSAT_print_assignment (CompSat * compsat, FILE * file)
{
  char buffer[LINE_WIDTH], str[20];
  int i, tmp, pos, len;

  assert (LINE_WIDTH > MAX_POS);

  pos = 0;

  pos = 0;

  for (i = 1; i <= compsat->max + 1; i++)
    {
      if (i <= compsat->max)
	{
	  tmp = get_assignment (compsat, i);
	  tmp = (tmp == FALSE) ? -i : i;
	}
      else
	tmp = 0;			/* SAT 2003 competition format */

      /* TODO: this could be slow.
       */
      sprintf (str, "%d", tmp);
      len = strlen (str);

      if (pos + len >= MAX_POS)
	{
	  buffer[pos] = 0;
	  fputs ("v ", file);		/* SAT 2003 competition format */
	  fputs (buffer, file);
	  fputc ('\n', file);
	  pos = 0;
	}

      if (pos)
	buffer[pos++] = ' ';

      strcpy (buffer + pos, str);
      pos += len;
    }

  if (pos)
    {
      fputs ("v ", file);		/* SAT 2003 competition format */
      fputs (buffer, file);
      fputc ('\n', file);
    }
}

/*------------------------------------------------------------------------*/


inline_static int
is_number (const char *str)
{
  const char *p;
  int res;

  p = str;

  if (*p == '-')
    p++;

  if (isdigit ((int) *p++))
    {
      while (isdigit ((int) *p))
	p++;

      res = !*p;
    }
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static char
character_class (char ch)
{
  if (ch >= '0' && ch <= '9')
    return '0';

  return ch;
}

/*------------------------------------------------------------------------*/

inline_static int
is_float (const char *str)
{
  const char *p;
  int digits;

  digits = 0;
  p = str;

  goto L1;

L1:				/* -?(0*.0+ | 0+) */

  switch (character_class (*p++))
    {
    case '-':
      goto L2;

    case '.':
      goto L3;

    case '0':
      goto L4;
    }

  return 0;

L2:				/* 0*.0+ | 0+ */

  switch (character_class (*p++))
    {
    case '0':
      goto L4;

    case '.':
      goto L3;
    }

  return 0;

L3:				/* 0+ */

  switch (character_class (*p++))
    {
    case '0':
      goto L5;
    }

  return 0;

L4:				/* 0*.0+ | 0* */

  switch (character_class (*p++))
    {
    case '0':
      goto L4;

    case '.':
      goto L3;

    case 0:
      return 1;
    }

  return 0;

L5:				/* 0* */

  switch (character_class (*p++))
    {
    case '0':
      goto L5;

    case 0:
      return 1;
    }

  return 0;
}

/*------------------------------------------------------------------------*/

COMPSAT_ArgvIterator *
COMPSAT_new_ArgvIterator (CompSat * compsat, int argc, char **argv)
{
  COMPSAT_ArgvIteratorInternal *internal;
  COMPSAT_ArgvIterator *res;

  internal = new (compsat, sizeof (*internal));
  internal->compsat = compsat;
  internal->argc = argc;
  internal->argv = argv;
  internal->pos = 1;
  internal->opt_buffer[0] = '-';

  res = new (compsat, sizeof (*res));
  res->internal = internal;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
COMPSAT_is_option_with_argument (const char *opt)
{
  assert (opt[0] == '-');
  assert (opt[2] == 0);

  switch (opt[1])
    {
    case 'l':
    case 'o':
    case 't':
    case 'd':
    case 'm':
#ifdef COMPSAT_WHITE
    case 'W':
#endif
      return 1;
    default:
      return 0;
    }
}

/*------------------------------------------------------------------------*/

int
COMPSAT_is_file_option (const char *opt)
{
  if (opt[0] == '-' && !opt[2])
    return (opt[1] == 'l' || opt[1] == 'o');

  return !strcmp (opt, "--divert");
}

/*------------------------------------------------------------------------*/

void
COMPSAT_report_invalid_option (CompSat * compsat,
			      COMPSAT_ArgvIterator * it, const char *err)
{
  FILE *file;

  file = COMPSAT_err_file (compsat);

  fprintf (file, "*** '%s", it->opt);
  if (it->arg)
    fprintf (file, "[ ]%s", it->arg);
  fprintf (file, "': %s\n", err);
  fflush (file);
}

/*------------------------------------------------------------------------*/

int
COMPSAT_has_more_elements_ArgvIterator (COMPSAT_ArgvIterator * it,
				       const char **err_msg)
{
  char *str;

  *err_msg = 0;

  if (it->internal->pos >= it->internal->argc)
    return 0;

  str = it->internal->argv[it->internal->pos++];
  it->opt = str;

  if (str[0] == '-' && str[1] && str[1] != '-')
    {
      if (str[2])
	{
	  it->arg = str + 2;
	  it->internal->opt_buffer[1] = str[1];
	  it->opt = it->internal->opt_buffer;
	}
      else if (!COMPSAT_is_option_with_argument (str))
	it->arg = 0;
      else if (it->internal->pos < it->internal->argc)
	it->arg = it->internal->argv[it->internal->pos++];
      else
	{
	  *err_msg = "argument is missing";
	  return 0;
	}
    }
  else
    it->arg = 0;

  return 1;
}

/*------------------------------------------------------------------------*/

void
COMPSAT_delete_ArgvIterator (COMPSAT_ArgvIterator * it)
{
  CompSat *compsat;

  compsat = it->internal->compsat;
  delete (compsat, it->internal, sizeof (*it->internal));
  delete (compsat, it, sizeof (*it));
}

/*------------------------------------------------------------------------*/

#ifdef COMPSAT_CHECK
#define CHECK_OPTION \
"  --check            enable expensive checks\n"
#else
#define CHECK_OPTION ""
#endif

#define USAGE \
"  -v                 increase verbosity level\n" \
"  -V                 set maximal verbosity\n" \
"  -o[ ]<out-file>    use <out-file> for printing (default stdout)\n" \
"  -l[ ]<log-file>    use <log-file> for logging (default stderr)\n" \
"  --divert           divert logging to <out-file>\n" \
CHECK_OPTION \
"\n" \
"  -d[ ]<num>         max decisions (default -1 = unrestricted)\n" \
"  -m[ ]<num>         max mega bytes (default -1 = unrestricted)\n" \
"  -t[ ]<num>         max seconds (default -1 = unrestricted)\n"

/*------------------------------------------------------------------------*/

int
COMPSAT_verbose_level (CompSat * compsat)
{
  return compsat->verbosity;
}

/*------------------------------------------------------------------------*/

const char *
COMPSAT_prefix (CompSat * compsat)
{
  (void) compsat;
  return COMPSAT_PREFIX;
}

/*------------------------------------------------------------------------*/

const char *
COMPSAT_usage (void)
{
  return USAGE;
}

/*------------------------------------------------------------------------*/

static const char *could_not_open_file = "could not open file";
static const char *argument_is_missing = "argument is missing";
static const char *argument_is_no_integer = "argument is no integer";
static const char *argument_is_no_float = "argument is no float";
static const char *invalid_option = "invalid option";

/*------------------------------------------------------------------------*/

const char *
COMPSAT_option (CompSat * compsat, const char *opt, const char *arg)
{
  double number;

  if (!strcmp (opt, "-o"))
    {
      if (!arg)
	return argument_is_missing;

      if (compsat->out)
	fclose (compsat->out);

      compsat->out = fopen (arg, "w");
      if (!compsat->out)
	return could_not_open_file;
    }
  else if (!strcmp (opt, "-l"))
    {
      if (!arg)
	return argument_is_missing;

      if (compsat->log)
	fclose (compsat->log);

      compsat->log = fopen (arg, "w");
      if (!compsat->out)
	return could_not_open_file;
    }
#ifdef COMPSAT_CHECK
  else if (!strcmp (opt, "--check"))
    {
      compsat->checker.enabled = 1;
    }
#endif
  else if (!strcmp (opt, "--divert"))
    {
      compsat->divert = 1;
    }
  else if (!strcmp (opt, "-v"))
    {
      if (compsat->verbosity < INT_MAX)
	compsat->verbosity++;
    }
  else if (!strcmp (opt, "-V"))
    {
      compsat->verbosity = INT_MAX;
    }
  else if (!strcmp (opt, "-m"))
    {
      if (!arg)
	return argument_is_missing;

      if (!is_float (arg))
	return argument_is_no_float;

      number = strtod (arg, NULL);
      if (number >= (LONG_MAX >> 20))
	return 0;

      if (number < 0)
	compsat->limits.bytes = -1;
      else
	compsat->limits.bytes = (number * (1 << 20));
    }
  else if (!strcmp (opt, "-d"))
    {
      if (!arg)
	return argument_is_missing;

      if (!is_number (arg))
	return argument_is_no_integer;

      number = strtod (arg, NULL);
      compsat->limits.decisions = (number < 0) ? -1 : number;
    }
  else if (!strcmp (opt, "-t"))
    {
      if (!arg)
	return argument_is_missing;

      if (!is_float (arg))
	return argument_is_no_float;

      number = strtod (arg, NULL);
      compsat->limits.time = (number < 0) ? -1 : number;

      /* TODO: do we want to add the already used up time here ? */
    }
  else if (opt[0] == '-')
    return invalid_option;
  else
    compsat->name = copy_string (compsat, opt);

  return 0;
}

/*------------------------------------------------------------------------*/

const char *
COMPSAT_id (void)
{
  return COMPSAT_ID;
}

/*------------------------------------------------------------------------*/

const char *
COMPSAT_copyright (void)
{
  return COMPSAT_COPYRIGHT;
}

/*------------------------------------------------------------------------*/

const char *
COMPSAT_version (void)
{
  return COMPSAT_VERSION;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

static void
white_white (CompSat * compsat)
{
  out (compsat, "this is the output of the white-box test case\n");
}

/*------------------------------------------------------------------------*/

static void
white_stack (CompSat * compsat)
{
  FastStack stack;
  int i;

  memset (&stack, 0, sizeof (stack));

  reset_FastStack (compsat, &stack, 0);

  for (i = 0; i < 2000; i++)
    push_FastStack (compsat, &stack, i);

  for (i = 0; i < 1000; i++)
    (void) pop_FastStack (compsat, &stack);

  reset_FastStack (compsat, &stack, 500);

  set_FastStack (compsat, &stack, 499, -499);
  set_FastStack (compsat, &stack, 500, -500);
  set_FastStack (compsat, &stack, 501, -501);
  set_FastStack (compsat, &stack, 511, -511);
  set_FastStack (compsat, &stack, 513, -513);

  while (!is_empty_FastStack (&stack))
    out (compsat, "%d\n", pop_FastStack (compsat, &stack));

  release_FastStack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_pointer0 (CompSat * compsat)
{
  PointerStack stack;
  int tmp;

  memset (&stack, 0, sizeof (stack));
  push_PointerStack (compsat, &stack, 0);
  push_PointerStack (compsat, &stack, 1);
  push_PointerStack (compsat, &stack, 2);
  push_PointerStack (compsat, &stack, 3);
  push_PointerStack (compsat, &stack, 4);
  {
    CursorPointerStack (&stack, it, 2, -1);
    while ((tmp = next_PointerStack (it)))
      out (compsat, "%d\n", tmp);
  }
  out (compsat, "---------------------------------\n");
  {
    CursorPointerStack (&stack, it, 1, 1);
    while ((tmp = next_PointerStack (it)) != 4)
      out (compsat, "%d\n", tmp);
  }
  out (compsat, "---------------------------------\n");
  {
    IteratorPointerStack (&stack, it, 1);
    while (!done_PointerStack (it))
      out (compsat, "%d\n", next_PointerStack (it));
  }
  out (compsat, "---------------------------------\n");
  {
    IteratorPointerStack (&stack, it, -1);
    while (!done_PointerStack (it))
      out (compsat, "%d\n", next_PointerStack (it));
  }
  out (compsat, "---------------------------------\n");
  reset_PointerStack (compsat, &stack, 0);
  push_PointerStack (compsat, &stack, 5);
  {
    IteratorPointerStack (&stack, it, 1);
    while (!done_PointerStack (it))
      out (compsat, "%d\n", next_PointerStack (it));
  }
  out (compsat, "---------------------------------\n");
  release_PointerStack (compsat, &stack);
  push_PointerStack (compsat, &stack, 6);
  {
    IteratorPointerStack (&stack, it, 1);
    while (!done_PointerStack (it))
      out (compsat, "%d\n", next_PointerStack (it));
  }
  release_PointerStack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_pointer1 (CompSat * compsat)
{
  PointerStack stack;
  void *s[2];
  int tmp;

  memset (&stack, 0, sizeof (stack));
  push_PointerStack (compsat, &stack, 0);
  push_PointerStack (compsat, &stack, 1);
  push_PointerStack (compsat, &stack, 2);

  {
    CursorPointerStack (&stack, it, 0, 1);
    save_CursorPointerStack (it, s[0]);
    (void) next_PointerStack (it);
    save_CursorPointerStack (it, s[1]);
    (void) next_PointerStack (it);
    swap_PointerStack (&stack, it, s[0], s[1]);
  }

  {
    IteratorPointerStack (&stack, it, 1);
    while (!done_PointerStack (it))
      out (compsat, "%d\n", next_PointerStack (it));
  }

  out (compsat, "---------------------------------\n");

  {
    CursorPointerStack (&stack, it, 2, -1);
    save_CursorPointerStack (it, s[0]);
    (void) next_PointerStack (it);
    save_CursorPointerStack (it, s[1]);
    tmp = next_PointerStack (it);
    assert (tmp == 0);
    swap_PointerStack (&stack, it, s[0], s[1]);
  }

  {
    IteratorPointerStack (&stack, it, 1);
    while (!done_PointerStack (it))
      out (compsat, "%d\n", next_PointerStack (it));
  }

  release_PointerStack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_compact (CompSat * compsat)
{
  CompactStack stack;
  memset (&stack, 0, sizeof (stack));
  enlarge_CompactStack (compsat, &stack);
  log (compsat, 0, "%d\n", size_CompactStack (&stack));
  release_CompactStack (compsat, &stack);
  log (compsat, 0, "%d\n", size_CompactStack (&stack));
  enlarge_CompactStack (compsat, &stack);
  log (compsat, 0, "%d\n", size_CompactStack (&stack));
  enlarge_CompactStack (compsat, &stack);
  log (compsat, 0, "%d\n", size_CompactStack (&stack));
  reset_CompactStack (compsat, &stack, 0);
  log (compsat, 0, "%d\n", size_CompactStack (&stack));
  enlarge_CompactStack (compsat, &stack);
  log (compsat, 0, "%d\n", size_CompactStack (&stack));
  enlarge_CompactStack (compsat, &stack);
  log (compsat, 0, "%d\n", size_CompactStack (&stack));
  release_CompactStack (compsat, &stack);
  log (compsat, 0, "%d\n", size_CompactStack (&stack));
}

/*------------------------------------------------------------------------*/

static void
white_compact0 (CompSat * compsat)
{
  CompactStack stack;
  int elem;

  memset (&stack, 0, sizeof (stack));
  push_CompactStack (compsat, &stack, 0);
  push_CompactStack (compsat, &stack, 1);
  push_CompactStack (compsat, &stack, 2);
  push_CompactStack (compsat, &stack, 3);
  push_CompactStack (compsat, &stack, 4);

  {
    IteratorCompactStack (&stack, it, 1);
    while (!done_CompactStack (it))
      {
	elem = next_CompactStack (it);
	log (compsat, 0, "%d %d\n", elem, size_CompactStack (&stack));
      }
  }

  log (compsat, 0, "------------------------------------\n");

  push_CompactStack (compsat, &stack, 5);

  {
    IteratorCompactStack (&stack, it, 1);
    while (!done_CompactStack (it))
      {
	elem = next_CompactStack (it);
	log (compsat, 0, "%d %d\n", elem, size_CompactStack (&stack));
      }
  }

  log (compsat, 0, "------------------------------------\n");

  reset_CompactStack (compsat, &stack, 0);

  push_CompactStack (compsat, &stack, 1000);
  push_CompactStack (compsat, &stack, 1001);
  push_CompactStack (compsat, &stack, 1002);
  push_CompactStack (compsat, &stack, 1003);
  push_CompactStack (compsat, &stack, 1004);
  push_CompactStack (compsat, &stack, 1005);
  push_CompactStack (compsat, &stack, 65536);
  push_CompactStack (compsat, &stack, 1007);
  push_CompactStack (compsat, &stack, 1008);
  push_CompactStack (compsat, &stack, 1009);

  {
    IteratorCompactStack (&stack, it, 1);
    while (!done_CompactStack (it))
      {
	elem = next_CompactStack (it);
	log (compsat, 0, "%d %d\n", elem, size_CompactStack (&stack));
      }
  }

  log (compsat, 0, "------------------------------------\n");

  push_CompactStack (compsat, &stack, 2000);
  push_CompactStack (compsat, &stack, 2001);
  push_CompactStack (compsat, &stack, 2002);
  push_CompactStack (compsat, &stack, 2003);

  {
    IteratorCompactStack (&stack, it, -1);
    while (!done_CompactStack (it))
      {
	elem = next_CompactStack (it);
	log (compsat, 0, "%d %d\n", elem, size_CompactStack (&stack));
      }
  }

  reset_CompactStack (compsat, &stack, 0);

  {
    IteratorCompactStack (&stack, it, -1);
    assert (done_CompactStack (it) || next_CompactStack (it));
  }
  {
    IteratorCompactStack (&stack, it, 1);
    assert (done_CompactStack (it) || next_CompactStack (it));
  }

  release_CompactStack (compsat, &stack);
  {
    IteratorCompactStack (&stack, it, 1);
    assert (done_CompactStack (it) || next_CompactStack (it));
  }
  {
    IteratorCompactStack (&stack, it, -1);
    assert (done_CompactStack (it) || next_CompactStack (it));
  }
}

/*------------------------------------------------------------------------*/

static void
white_compact1_line (CompSat * compsat, CompactStack * stack)
{
  out (compsat, "-- count=%d -----------\n", count_CompactStack (stack));
}

/*------------------------------------------------------------------------*/

static void
white_compact1 (CompSat * compsat)
{
  CompactStack stack;
  int tmp, i;

  memset (&stack, 0, sizeof (stack));
  for (i = 0; i < 2; i++)
    {
      push_CompactStack (compsat, &stack, 0);
      push_CompactStack (compsat, &stack, 1);
      push_CompactStack (compsat, &stack, 2);
      push_CompactStack (compsat, &stack, 3);
      push_CompactStack (compsat, &stack, 4);
      if (i)
	push_CompactStack (compsat, &stack, 1 << 16);

      {
	CursorCompactStack (&stack, it, 2, -1);
	while ((tmp = next_CompactStack (it)))
	  out (compsat, "%d\n", tmp);
      }
      white_compact1_line (compsat, &stack);
      {
	CursorCompactStack (&stack, it, 1, 1);
	while ((tmp = next_CompactStack (it)) != 4)
	  out (compsat, "%d\n", tmp);
      }
      white_compact1_line (compsat, &stack);
      {
	IteratorCompactStack (&stack, it, 1);
	while (!done_CompactStack (it))
	  out (compsat, "%d\n", next_CompactStack (it));
      }
      white_compact1_line (compsat, &stack);
      {
	IteratorCompactStack (&stack, it, -1);
	while (!done_CompactStack (it))
	  out (compsat, "%d\n", next_CompactStack (it));
      }
      white_compact1_line (compsat, &stack);
      reset_CompactStack (compsat, &stack, 0);
      push_CompactStack (compsat, &stack, 5);
      {
	IteratorCompactStack (&stack, it, 1);
	while (!done_CompactStack (it))
	  out (compsat, "%d\n", next_CompactStack (it));
      }
      white_compact1_line (compsat, &stack);
      release_CompactStack (compsat, &stack);
      push_CompactStack (compsat, &stack, 6);
      {
	IteratorCompactStack (&stack, it, 1);
	while (!done_CompactStack (it))
	  out (compsat, "%d\n", next_CompactStack (it));
      }
      release_CompactStack (compsat, &stack);
      if (!i)
	white_compact1_line (compsat, &stack);
    }
}

/*------------------------------------------------------------------------*/

static void
white_compact2 (CompSat * compsat)
{
  CompactStack stack;

  memset (&stack, 0, sizeof (stack));
  push_CompactStack (compsat, &stack, 0);
  push_CompactStack (compsat, &stack, 1);
  push_CompactStack (compsat, &stack, 2);
  push_CompactStack (compsat, &stack, 3);
  push_CompactStack (compsat, &stack, 4);

  out (compsat, "%d\n", get_CompactStack (&stack, 17));
  out (compsat, "%d\n", get_CompactStack (&stack, 1));
  out (compsat, "%d\n", get_CompactStack (&stack, 1 << 16));
  out (compsat, "%d\n", get_CompactStack (&stack, 4));
  push_CompactStack (compsat, &stack, 1 << 16);
  out (compsat, "%d\n", get_CompactStack (&stack, 17));
  out (compsat, "%d\n", get_CompactStack (&stack, 1));
  out (compsat, "%d\n", get_CompactStack (&stack, 1 << 16));
  out (compsat, "%d\n", get_CompactStack (&stack, 4));
  out (compsat, "%d\n", get_CompactStack (&stack, 5));

  release_CompactStack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_compact3 (CompSat * compsat)
{
  CompactStack stack;
  void *s[2];
  int tmp;

  memset (&stack, 0, sizeof (stack));
  push_CompactStack (compsat, &stack, 0);
  push_CompactStack (compsat, &stack, 1);
  push_CompactStack (compsat, &stack, 2);

  {
    CursorCompactStack (&stack, it, 0, 1);
    save_CursorCompactStack (it, s[0]);
    (void) next_CompactStack (it);
    save_CursorCompactStack (it, s[1]);
    (void) next_CompactStack (it);
    swap_CompactStack (&stack, it, s[0], s[1]);
  }

  {
    IteratorCompactStack (&stack, it, 1);
    while (!done_CompactStack (it))
      out (compsat, "%d\n", next_CompactStack (it));
  }

  out (compsat, "---------------------------------\n");

  {
    CursorCompactStack (&stack, it, 2, -1);
    save_CursorCompactStack (it, s[0]);
    (void) next_CompactStack (it);
    save_CursorCompactStack (it, s[1]);
    tmp = next_CompactStack (it);
    assert (tmp == 0);
    swap_CompactStack (&stack, it, s[0], s[1]);
  }

  {
    IteratorCompactStack (&stack, it, 1);
    while (!done_CompactStack (it))
      out (compsat, "%d\n", next_CompactStack (it));
  }

  push_CompactStack (compsat, &stack, 1 << 16);

  {
    CursorCompactStack (&stack, it, 3, -1);
    save_CursorCompactStack (it, s[0]);
    (void) next_CompactStack (it);
    save_CursorCompactStack (it, s[1]);
    tmp = next_CompactStack (it);
    assert (tmp == 0);
    swap_CompactStack (&stack, it, s[0], s[1]);
  }

  out (compsat, "---------------------------------\n");

  {
    IteratorCompactStack (&stack, it, 1);
    while (!done_CompactStack (it))
      out (compsat, "%d\n", next_CompactStack (it));
  }

  release_CompactStack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_compact4 (CompSat * compsat)
{
  CompactStack stack;

  memset (&stack, 0, sizeof (stack));

  push_CompactStack (compsat, &stack, 32767);
  push_CompactStack (compsat, &stack, -32768);
  push_CompactStack (compsat, &stack, 32768);

  {
    IteratorCompactStack (&stack, it, 1);
    while (!done_CompactStack (it))
      out (compsat, "%d\n", next_CompactStack (it));
  }

  release_CompactStack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

#define N_WHITE_QUEUE 40
#define M_WHITE_QUEUE 100

static void
white_queue (CompSat * compsat)
{
  void *ptr[M_WHITE_QUEUE];
  Queue queue;
  int i, j, k;

  k = 1;
  memset (ptr, 0, sizeof (ptr));

  for (i = 0; i < N_WHITE_QUEUE; i++)
    {
      memset (&queue, 0, sizeof (queue));
      reset_Queue (compsat, &queue);

      delete (compsat, ptr[k], ptr[k] ? k + 1 : 0);
      ptr[k] = new (compsat, k + 1);
      if (++k >= M_WHITE_QUEUE)
	k = 0;

      for (j = 0; j < N_WHITE_QUEUE / 2; j++)
	enqueue (compsat, &queue, j);

      for (j = 0; j < i && j < N_WHITE_QUEUE / 2; j++)
	out (compsat, "%dA%d\n", i, dequeue (compsat, &queue));

      delete (compsat, ptr[k], ptr[k] ? k + 1 : 0);
      ptr[k] = new (compsat, k + 1);
      if (++k >= M_WHITE_QUEUE)
	k = 0;

      for (j = N_WHITE_QUEUE / 2; j < N_WHITE_QUEUE; j++)
	enqueue (compsat, &queue, j);

      while (!is_empty_Queue (&queue))
	out (compsat, "%dB%d\n", i, dequeue (compsat, &queue));

      release_Queue (compsat, &queue);
    }

  for (k = 0; k < M_WHITE_QUEUE; k++)
    delete (compsat, ptr[k], ptr[k] ? k + 1 : 0);
}

/*------------------------------------------------------------------------*/

#define N_WHITE_BITMAP 1000

static void
white_bitmap (CompSat * compsat)
{
  int i;

#ifdef USE_BITMAP
  FastStack bitmap;

  memset (&bitmap, 0, sizeof (bitmap));

  for (i = 0; i < N_WHITE_BITMAP; i += 3)
    set_bit (compsat, &bitmap, i);

  for (i = 0; i < N_WHITE_BITMAP; i += 2)
    clr_bit (compsat, &bitmap, i);

  for (i = 0; i < N_WHITE_BITMAP; i++)
    if (get_bit (compsat, &bitmap, i))
      out (compsat, "%d\n", i);

  release_FastStack (compsat, &bitmap);
#else
  for (i = 0; i < N_WHITE_BITMAP; i++)
    if (!(i % 3) && (i % 2))
      out (compsat, "%d\n", i);
#endif
}

/*------------------------------------------------------------------------*/

#define N_WHITE_PPLONG 1000
#define M_WHITE_PPLONG 10

static void
white_pplong (CompSat * compsat)
{
  char name[200];
  int i, j, r, k;
  FILE *file;
  char ch;

  sprintf (name, "/tmp/white-pplong-%u", (unsigned) getpid ());
  file = fopen (name, "w");
  if (!file)
    return;

  fprintf (file, "p cnf %d %d\n", N_WHITE_PPLONG - 1, M_WHITE_PPLONG);

  r = 13;

  for (k = 0; k < M_WHITE_PPLONG; k++)
    {
      for (i = 1; i <= N_WHITE_PPLONG; i++)
	{
	  if (i < N_WHITE_PPLONG)
	    {
	      fprintf (file, "%d", (!(r % 2) ? -1 : 1) * i);
	      r = (101 * r + 17) % 113;
	    }
	  else
	    fputc ('0', file);

	  assert (r >= 0);

	  for (j = 0; j <= (i % 5); j++)
	    {
	      if (!((i + j) % 17))
		ch = '\t';
	      else if (!((i + j) % 31))
		ch = '\n';
	      else
		ch = ' ';

	      fputc (ch, file);
	    }
	}
    }

  fclose (file);

  COMPSAT_read (compsat, name);
  if (compsat->error)
    err (compsat, "%s", compsat->error);
  else
    out (compsat, "%d %d\n", compsat->max, count_clauses (compsat));

  unlink (name);
}

/*------------------------------------------------------------------------*/

#define N_WHITE_LARGE ((1 << 20) - 1)

static void
white_large (CompSat * compsat)
{
  char name[200];
  FILE *file;

  sprintf (name, "/tmp/white-large-%u", (unsigned) getpid ());
  file = fopen (name, "w");
  if (!file)
    return;

  fprintf (file, "p cnf %d 1\n%d 0\n", N_WHITE_LARGE, N_WHITE_LARGE);
  fclose (file);

  COMPSAT_read (compsat, name);
  if (compsat->error)
    err (compsat, "%s", compsat->error);
  else
    out (compsat, "%d %d\n", compsat->max, count_clauses (compsat));

  unlink (name);
}

/*------------------------------------------------------------------------*/

static void
white_memout (CompSat * compsat)
{
  int res;

#ifdef COMPSAT_WHITE
  void *ptr;

  ptr = new (compsat, 13);
  compsat->force_out_of_memory = 1;
  res = !new (compsat, 17);
  if (res)
    {
      res = !resize (compsat, ptr, 13, 200);
      if (res)
	delete (compsat, ptr, 13);
    }
#else
  res = 0;
#endif

  out (compsat, "%d\n", res);
}

/*------------------------------------------------------------------------*/

static void
white_memzero (CompSat * compsat)
{
  out (compsat, "%.0f\n", (double) (unsigned long) new (compsat, 0));
  out (compsat, "%.0f\n",
       (double) (unsigned long) resize (compsat, new (compsat, 17), 17, 0));
}

/*------------------------------------------------------------------------*/

static void
white_isfloat_aux (CompSat * compsat, const char *str)
{
  out (compsat, "\"%s\" %d\n", str, is_float (str));
}

/*------------------------------------------------------------------------*/

static void
white_isfloat (CompSat * compsat)
{
  white_isfloat_aux (compsat, "0");
  white_isfloat_aux (compsat, ".0");
  white_isfloat_aux (compsat, "3.1415");
  white_isfloat_aux (compsat, "-1");
  white_isfloat_aux (compsat, "-1.23");
  white_isfloat_aux (compsat, "-.2");
  white_isfloat_aux (compsat, "47.11");
  white_isfloat_aux (compsat, "-47.11");

  white_isfloat_aux (compsat, "");
  white_isfloat_aux (compsat, "a");
  white_isfloat_aux (compsat, ".");
  white_isfloat_aux (compsat, "0.");
  white_isfloat_aux (compsat, "-a");
  white_isfloat_aux (compsat, "0a");
  white_isfloat_aux (compsat, "0.0a");
}

/*------------------------------------------------------------------------*/

static void
white_prass (CompSat * compsat)
{
  int i;

  enlarge (compsat, 1000);
  for (i = 1; i <= compsat->max; i += 2)
    {
      set_assignment (compsat, i, TRUE);
      set_assignment (compsat, i, FALSE);
    }

  COMPSAT_print_assignment (compsat, out_file (compsat));
}

/*------------------------------------------------------------------------*/

static void
white_dimacs0 (CompSat * compsat)
{
  FILE *file = log_file (compsat);
  const char *p;

  for (p = DIMACS_id (); *p; p++)
    {
      switch (*p)
	{
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case '.':
	case '$':
	  break;

	default:
	  fputc (*p, file);
	  break;
	}
    }

  fputc ('\n', file);
}

/*------------------------------------------------------------------------*/

static void
white_log_from_file (CompSat * compsat, const char *name)
{
  FILE *file;
  int ch;

  file = fopen (name, "r");
  if (file)
    {
      while ((ch = fgetc (file)) != EOF)
	log (compsat, 0, "%c\n", ch);

      fclose (file);
    }
}

/*------------------------------------------------------------------------*/

#define LOG_STR "/tmp/compsat-white-log"

static void
white_log (CompSat * compsat)
{
  CompSat *other;
  other = COMPSAT_new ();
  COMPSAT_option (other, "-l", LOG_STR);
  log (other, 0, "logged\n");
  COMPSAT_delete (other);
  white_log_from_file (compsat, LOG_STR);
  unlink (LOG_STR);
}

/*------------------------------------------------------------------------*/

#define OUT_STR "/tmp/compsat-white-out"

static void
white_out (CompSat * compsat)
{
  CompSat *other;
  other = COMPSAT_new ();
  COMPSAT_option (other, "-o", OUT_STR);
  out (other, "printed\n");
  COMPSAT_option (other, "--divert", 0);
  log (other, 0, "logged again\n");
  COMPSAT_delete (other);
  white_log_from_file (compsat, OUT_STR);
  unlink (OUT_STR);
}

/*------------------------------------------------------------------------*/

static void
white_cmp (CompSat * compsat)
{
  int res, clause[3];

  clause[0] = 1;
  clause[1] = 2;
  clause[2] = 0;
  COMPSAT_add (compsat, clause);
  clause[0] = -1;
  clause[1] = 0;
  COMPSAT_add (compsat, clause);

  res = 1;

  if (res)
    res = (cmp (compsat, 1, 1) == 0);
  if (res)
    res = (cmp (compsat, -1, -1) == 0);

  if (res)
    res = (cmp (compsat, 1, -1) < 0);
  if (res)
    res = (cmp (compsat, -1, 1) > 0);

  if (res)
    res = (cmp (compsat, 1, 2) > 0);
  if (res)
    res = (cmp (compsat, 2, 1) < 0);

  set_score (compsat, 2, 1);

  if (res)
    res = (cmp (compsat, 1, 2) < 0);
  if (res)
    res = (cmp (compsat, 2, 1) > 0);

  set_score (compsat, 1, 1);
  inc_ref (compsat, 1);

  if (res)
    res = (cmp (compsat, 1, 2) > 0);
  if (res)
    res = (cmp (compsat, 2, 1) < 0);

  log (compsat, 0, "%d\n", res);
}

/*------------------------------------------------------------------------*/

#define N_SORT 111 * 111	/* has to be square number */

static void
white_sort (CompSat * compsat)
{
  int *clause, i, j, delta, count, sorted;

  clause = new (compsat, sizeof (clause[0]) * (N_SORT + 1));
  for (delta = 0; delta * delta < N_SORT; delta++)
    ;
  assert (delta * delta == N_SORT);
  count = 1;
  for (i = 0; i < delta; i++)
    for (j = 0; j < delta; j++)
      clause[j * delta + i] = count++;
  clause[N_SORT] = 0;
  COMPSAT_add (compsat, clause);

  hypersort (compsat, clause, N_SORT);

  sorted = 1;
  for (i = 0; sorted && i < N_SORT - 1; i++)
    sorted = (cmp (compsat, clause[i], clause[i + 1]) > 0);

  delete (compsat, clause, sizeof (clause[0]) * (N_SORT + 1));

  log (compsat, 0, "%s\n", sorted ? "sorted" : "unsorted");
}

/*------------------------------------------------------------------------*/

typedef void (*TestCase) (CompSat *);

#define TC(pattern) \
do { if(!strcmp(# pattern, name)) return white_ ## pattern; } while(0)

static TestCase
get_white_box_test_case (const char *name)
{
  TC (white);
  TC (stack);
  TC (pointer0);
  TC (pointer1);
  TC (compact);
  TC (compact0);
  TC (compact1);
  TC (compact2);
  TC (compact3);
  TC (compact4);
  TC (queue);
  TC (bitmap);
  TC (pplong);
  TC (large);
  TC (memout);
  TC (memzero);
  TC (isfloat);
  TC (prass);
  TC (dimacs0);
  TC (log);
  TC (out);
  TC (cmp);
  TC (sort);

  return 0;
}

/*------------------------------------------------------------------------*/

int
COMPSAT_is_white_box_test_case (const char *name)
{
  return get_white_box_test_case (name) != 0;
}

/*------------------------------------------------------------------------*/

void
COMPSAT_run_white_box_test_case (CompSat * compsat, const char *name)
{
  TestCase tc;

  tc = get_white_box_test_case (name);
  assert (tc);
  tc (compsat);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#if 0
/*------------------------------------------------------------------------*/
/* This section contains some external visible code for manually checking
 * the efficiency of the compiler in handling the compact stack iterators by
 * reviewing the produced assembler code or by benchmarking it with
 * 'fillcompactstack.c'.
 */

int
count_forward_CompactStack (CompactStack * stack)
{
  IteratorCompactStack (stack, it, 1);
  int res;

  res = 0;
  while (!done_CompactStack (it))
    res += next_CompactStack (it);

  return res;
}

/*------------------------------------------------------------------------*/

int
count_backward_CompactStack (CompactStack * stack)
{
  IteratorCompactStack (stack, it, -1);
  int res;

  res = 0;
  while (!done_CompactStack (it))
    res += next_CompactStack (it);

  return res;
}

/*------------------------------------------------------------------------*/

void
fill_CompactStack (CompSat * compsat, CompactStack * stack, int n)
{
  int i, tmp;

  tmp = 4;
  for (i = 0; i < n; i++)
    {
      tmp = 7000 + (13 * tmp + 1) % 17;
      push_CompactStack (compsat, stack, n);
    }
}

/*------------------------------------------------------------------------*/
#endif
