/*------------------------------------------------------------------------*/

const char *
DIMACS_id (void)
{
  return "$Id: dimacs.c,v 1.20 2002/03/22 14:57:15 biere Exp $";
}

/*------------------------------------------------------------------------*/

#include "dimacs.h"

/*------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>

/*------------------------------------------------------------------------*/

struct Dimacs
{
  int count;			/* number literals on 'literals' stack */
  int size;			/* size of the 'literals' stack */
  int *literals;		/* stack to save the literals of a
				 * clause */
  int clauses;			/* specified number of clauses */
  int max;			/* specified maximal number of variables */
  int lineno;			/* current lineno */
  char *error;			/* holds error message if error occured */
  char *name;			/* name of the input file */
  char *start;			/* start of the input file in memory */
  char *pos;			/* current position in the input file */
  char *end;			/* end of the input file in memory */
  size_t error_len;		/* size memory allocated for 'error' */
  DimacsMem *mem;		/* external memory manager */
};

/*------------------------------------------------------------------------*/
#if 0
/*
 * Only needed if we implement memory mapped file IO
 */
static void
adjust_bytes (Dimacs * dimacs, size_t delta)
{
  if (dimacs->mem && dimacs->mem->bytes_ptr)
    {
      *dimacs->mem->bytes_ptr += delta;

      if (dimacs->mem->max_bytes_ptr &&
	  *dimacs->mem->max_bytes_ptr > *dimacs->mem->bytes_ptr)
	*dimacs->mem->max_bytes_ptr = *dimacs->mem->bytes_ptr;
    }
}

#endif
/*------------------------------------------------------------------------*/

static void *
Dimacs_new (Dimacs * dimacs, size_t bytes)
{
  void *res;

  if (!bytes)
    res = 0;
  else
    {
      if (dimacs->mem && dimacs->mem->new)
	res = (*dimacs->mem->new) (dimacs->mem->mgr, bytes);
      else
	res = malloc (bytes);

      if (!res)
	{
	  fprintf (stderr,
		   "*** Dimacs_new(%lu): out of memory\n",
		   (unsigned long) bytes);

	  fflush (stderr);
	  abort ();
	}
    }

  return res;
}

/*------------------------------------------------------------------------*/

static void
Dimacs_delete (Dimacs * dimacs, void *ptr, size_t bytes)
{
  if (ptr)
    {
      if (dimacs->mem && dimacs->mem->delete)
	(*dimacs->mem->delete) (dimacs->mem->mgr, ptr, bytes);
      else
	free (ptr);
    }
}

/*------------------------------------------------------------------------*/

static int
file_size (const char *name)
{
  struct stat buf;
  int res;

  if (stat (name, &buf))
    return -1;

  res = buf.st_size;

  return res;
}

/*------------------------------------------------------------------------*/

static int
open_file (Dimacs * parser, const char *name)
{
  int size, count;
  FILE *file;

  file = fopen (name, "r");
  if (!file)
    return 0;

  size = file_size (name);
  count = 0;

  if (size > 0)
    {
      parser->start = (char *) Dimacs_new (parser, size);
      count = (int) fread (parser->start, 1, (size_t) size, file);
      parser->pos = parser->start;
      parser->end = parser->start + size;
    }
  fclose (file);

  return count == size;
}

/*------------------------------------------------------------------------*/

static int *
parse_error (Dimacs * parser, const char *msg)
{
  char *p, ch;

  Dimacs_delete (parser, parser->error, parser->error_len);
  parser->error_len = strlen (msg) + strlen (parser->name) + 100;
  parser->error = (char *) Dimacs_new (parser, parser->error_len);

  sprintf (parser->error,
	   "parse error:\n%s:%d: %s", parser->name, parser->lineno, msg);

  if (parser->pos < parser->end)
    {
      strcat (parser->error, " at ");

      p = parser->error + strlen (parser->error);
      ch = *parser->pos;

      if (ch == '\n')
	sprintf (p, "'\\n'");
      else if (ch == '\t')
	sprintf (p, "'\\t'");
      else if (isprint ((int) ch))
	sprintf (p, "'%c'", ch);
      else
	sprintf (p, "0x%x", ch);
    }
  return 0;
}

/*------------------------------------------------------------------------*/

Dimacs *
DIMACS_new (const char *name, DimacsMem * mem)
{
  Dimacs *res, dummy;

  memset (&dummy, 0, sizeof (dummy));
  dummy.mem = mem;
  res = (Dimacs *) Dimacs_new (&dummy, sizeof (*res));
  memset (res, 0, sizeof (*res));
  res->mem = mem;
  res->size = 1024;
  res->literals = (int *) Dimacs_new (res, sizeof (int) * res->size);
  res->name = (char *) Dimacs_new (res, strlen (name) + 1);
  strcpy (res->name, name);	/* save for error message */
  res->clauses = -1;		/* 'p cnf ...' spec not found sofar */
  res->max = -1;

  if (open_file (res, name))
    res->lineno = 1;
  else
    parse_error (res, "could not open");

  return res;
}

/*------------------------------------------------------------------------*/

void
DIMACS_delete (Dimacs * parser)
{
  Dimacs dummy;

  Dimacs_delete (parser, parser->error, parser->error_len);
  Dimacs_delete (parser, parser->name, strlen (parser->name) + 1);
  Dimacs_delete (parser, parser->literals, parser->size * sizeof (int));
  Dimacs_delete (parser, parser->start, parser->end - parser->start);

  dummy.mem = parser->mem;
  Dimacs_delete (&dummy, parser, sizeof (*parser));
}

/*------------------------------------------------------------------------*/

const char *
DIMACS_error (Dimacs * parser)
{
  return parser->error;
}

/*------------------------------------------------------------------------*/

static void
enlarge (Dimacs * parser)
{
  int old_size, new_size, *new_literals;
  size_t old_bytes, new_bytes;

  old_size = parser->size;
  old_bytes = old_size * sizeof (int);

  new_size = 2 * old_size;
  new_bytes = new_size * sizeof (int);

  new_literals = Dimacs_new (parser, new_bytes);
  memcpy (new_literals, parser->literals, old_bytes);
  Dimacs_delete (parser, parser->literals, old_bytes);

  parser->size = new_size;
  parser->literals = new_literals;
}

/*------------------------------------------------------------------------*/

static void
insert (Dimacs * parser, int lit)
{
  if (parser->count >= parser->size)
    enlarge (parser);

  parser->literals[parser->count++] = lit;
}

/*------------------------------------------------------------------------*/

static int
space (Dimacs * parser)
{
  int found, ch;

  found = 0;

  while (parser->pos < parser->end &&
	 ((ch = *parser->pos) == ' ' || ch == '\t'))
    {
      parser->pos++;
      found = 1;
    }

  return found;
}

/*------------------------------------------------------------------------*/

static int
read_number (Dimacs * parser)
{
  int res, ch;

  res = 0;
  while (parser->pos < parser->end && isdigit ((int) (ch = *parser->pos)))
    {
      res = 10 * res + (ch - '0');
      parser->pos++;
    }

  return res;
}

/*------------------------------------------------------------------------*/

static int
number (Dimacs * parser)
{
  if (parser->pos >= parser->end || !isdigit ((int) *parser->pos))
    return -1;

  return read_number (parser);
}

/*------------------------------------------------------------------------*/

static int
match (Dimacs * parser, char ch)
{
  char buffer[20];

  if (parser->pos >= parser->end || *parser->pos != ch)
    {
      sprintf (buffer, "expected '%c'", ch);
      (void) parse_error (parser, buffer);
      return 0;
    }
  else
    {
      parser->pos++;
      return 1;
    }
}

/*------------------------------------------------------------------------*/

const int *
DIMACS_read (Dimacs * parser)
{
  int lit, sign;
  char ch;

  if (parser->error)
    return 0;

  parser->count = 0;

  while (parser->pos < parser->end)
    {
      ch = *parser->pos;

      if (isspace ((int) ch))
	{
	  if (ch == '\n')
	    parser->lineno++;

	  parser->pos++;
	}
      else if (ch == 'p')
	{
	  if (parser->clauses >= 0)
	    return parse_error (parser, "'p ...' spec found twice");

	  parser->pos++;
	  if (!space (parser))
	    return parse_error (parser, "expected white space after 'p'");

	  if (!match (parser, 'c') ||
	      !match (parser, 'n') || !match (parser, 'f'))
	    return 0;

	  if (!space (parser))
	    return parse_error (parser, "expected white space after 'cnf'");

	  if ((parser->max = number (parser)) < 0)
	    return parse_error (parser, "number of variables missing");

	  if (!space (parser))
	    return parse_error (parser, "expected white space after number");

	  if ((parser->clauses = number (parser)) < 0)
	    return parse_error (parser, "number of clauses missing");

	  continue;
	}
      else if (ch == 'c')
	{
	  parser->pos++;

	  while (parser->pos < parser->end && (ch = *parser->pos++) != '\n');

	  if (ch == '\n')
	    parser->lineno++;
	}
      else
	{

	  if (ch == '-')
	    {
	      sign = -1;
	      parser->pos++;

	      if (parser->pos >= parser->end ||
		  !isdigit ((int) (ch = *parser->pos)))
		return parse_error (parser, "expected digit after '-'");
	    }
	  else
	    sign = 1;

	  if (isdigit ((int) ch))
	    {
	      lit = read_number (parser);
	      lit *= sign;
	      insert (parser, lit);

	      if (lit)
		continue;
	      else
		return parser->literals;
	    }
	  else
	    return parse_error (parser, "invalid character");
	}
    }

  if (parser->count)
    return parse_error (parser, "clause without terminating '0'");

  if (parser->clauses < 0)
    return parse_error (parser, "'p cnf ...' missing");

  return 0;
}

/*------------------------------------------------------------------------*/

int
DIMACS_specified_max_variable (Dimacs * dimacs)
{
  return dimacs->max;
}

/*------------------------------------------------------------------------*/

int
DIMACS_specified_num_clauses (Dimacs * dimacs)
{
  return dimacs->clauses;
}
