#include "compsat.h"
#include "config.h"

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>

/*------------------------------------------------------------------------*/

int COMPSAT_set_leak_hook (CompSat *, size_t * leaked_ptr);
int COMPSAT_main (CompSat *, int argc, char **argv);

/*------------------------------------------------------------------------*/

typedef struct NameWithPriority NameWithPriority;

struct NameWithPriority
{
  char *name;
  int priority;
};

/*------------------------------------------------------------------------*/

static int verbosity = 0;
static int fast_test_cases_only = 0;
size_t leaked_bytes;

/*------------------------------------------------------------------------*/

static int
erase (void)
{
  const char * term;

  if(verbosity || !isatty(1))
    return 0;

  term = getenv ("TERM");
  if(!term)
    return 0;

  if(!strcmp(term, "dumb") || !strcmp(term, "emacs"))
    return 0;

  return 1;
}

/*------------------------------------------------------------------------*/

static int
has_opt_prefix (const char *name)
{
  const char *p;
  int found;

  p = name;
  found = 0;
  while (!found && *p)
    {
      if (*p++ == '.' &&
	  *p++ == 'o' && *p++ == 'p' && *p++ == 't' && *p++ == 's')
	{
	  if (!*p)
	    found = 1;
	}
    }

  return found;
}

/*------------------------------------------------------------------------*/

static char *
base (const char *name)
{
  char *res;
  int len;

  len = strlen (name);
  res = (char *) malloc (len + 1);
  strcpy (res, name);
  res[len - 5] = 0;

  return res;
}

/*------------------------------------------------------------------------*/

static const char *slow_test_cases[] = {
  "sort.opts",
  "pplong.opts",
  "prime169.opts",
  "prime121.opts",
  "prime289.opts",
  "prime361.opts",
  "prime529.opts",
  "prime841.opts",
  "prime961.opts",
  "prime1369.opts",
  "prime1681.opts",
  "prime1849.opts",
  "prime2209.opts",
  "add8.opts",
  "add16.opts"
};

/*------------------------------------------------------------------------*/

static int
is_slow (const char *name)
{
  int i;

  for (i = 0; i < sizeof (slow_test_cases) / sizeof (slow_test_cases[0]); i++)
    if (!strcmp (name, slow_test_cases[i]))
      return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

static int
match (const char *name, const char *pattern)
{
  const char *p, *q;

  for (p = name, q = pattern; *q && *q == *p; p++, q++)
    ;

  return !*q;
}

/*------------------------------------------------------------------------*/

static void
line (void)
{
  int i;

  for (i = 0; i < 77; i++)
    fputc ('-', stdout);

  fputc ('\n', stdout);
}

/*------------------------------------------------------------------------*/

static char *
cons (char *ptr, const char *suffix)
{
  ptr = (char *) realloc (ptr, strlen (ptr) + strlen (suffix) + 1);
  strcat (ptr, suffix);

  return ptr;
}

/*------------------------------------------------------------------------*/

#define FLAT_ARGV_SIZE 2000
#define ARGV_SIZE 20

/*------------------------------------------------------------------------*/

static void
parse_opts (const char *base, int check,
	    int *argc_ptr, char ***argv_ptr, char **flat_argv_ptr)
{
  char **argv, *flat_argv, *tmp, *arg;
  int i, ch, argc;
  FILE *file;

  flat_argv = (char *) malloc (FLAT_ARGV_SIZE);
  sprintf (flat_argv, "log/%s.opts", base);
  file = fopen (flat_argv, "r");
  i = 0;
  while ((ch = fgetc (file)) != EOF && ch != '\n')
    flat_argv[i++] = ch;
  fclose (file);
  flat_argv[i] = 0;

  assert (strlen (flat_argv) + 1 < FLAT_ARGV_SIZE);

#ifdef COMPSAT_CHECK
  if (check)
    flat_argv = cons (flat_argv, " --check");
#else
  assert (!check);
#endif

  flat_argv = cons (flat_argv, " --divert -olog/");
  flat_argv = cons (flat_argv, base);
  flat_argv = cons (flat_argv, ".log");

  tmp = strdup (flat_argv);
  argv = (char **) malloc (ARGV_SIZE * sizeof (char *));
  argv[0] = strdup ("./compsat");
  argv[1] = strtok (tmp, " ");
  argc = 2;
  while ((arg = strtok (0, " ")))
    argv[argc++] = strdup (arg);

  *argc_ptr = argc;
  *argv_ptr = argv;
  *flat_argv_ptr = flat_argv;
}

/*------------------------------------------------------------------------*/

static void
overwrite (char ch)
{
  int i;

  for (i = 0; i < 70; i++)
    fputc (ch, stdout);
}

/*------------------------------------------------------------------------*/

static int
next_char (FILE * file, int last_char)
{
  int ch;

  ch = fgetc (file);

  while (ch == 'c' && last_char == '\n')
    {
      while ((ch = fgetc (file)) != EOF && ch != '\n')
	;

      if (ch != EOF)
	{
	  last_char = '\n';
	  ch = fgetc (file);
	}
    }

  return ch;
}

/*------------------------------------------------------------------------*/
/* Compare two files for differences.  The result is '1' iff no differences
 * were encountered and both files could be accessed.  If one file could not
 * be openend, then '0' is returned.
 */
static int
cmp_files (char *name0, char *name1)
{
  int res, ch[2], last_char[2], done;
  FILE *f[2];

  last_char[0] = last_char [1] = '\n';

  f[0] = fopen (name0, "r");
  if (f[0])
    {
      f[1] = fopen (name1, "r");
      if (f[1])
	{
	  do
	    {
	      ch[0] = next_char (f[0], last_char[0]);
	      ch[1] = next_char (f[1], last_char[1]);
	      res = (ch[0] == ch[1]);
	      last_char[0] = ch[0];
	      last_char[1] = ch[1];
	      done = (ch[0] == EOF);
	    }
	  while (res && !done);
	  fclose (f[1]);
	}
      else
	res = 0;
      fclose (f[0]);
    }
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

static void
die (const char *fmt, ...)
{
  va_list ap;

  fprintf (stderr, "*** test: ");
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

/*------------------------------------------------------------------------*/

static int
priority (const char *name)
{
  char *path;
  FILE *file;
  int res;

  path = (char *) malloc (strlen ("log/.priority") + strlen (name) + 1);
  sprintf (path, "log/%s.priority", name);
  file = fopen (path, "r");
  if (file)
    {
      fscanf (file, "%d", &res);
      fclose (file);
    }
  else
    die ("could not read '%s' (see 'log/order')", path);

  free (path);

  return res;
}

/*------------------------------------------------------------------------*/

static void
print_args (int size, const char *argv)
{
  int i;

  for (i = 0; i < size; i++)
    {
      if (argv[i + 0] == '-' && argv[i + 1] == '-' && argv[i + 2] == 'd' &&
	  argv[i + 3] == 'i' && argv[i + 4] == 'v' && argv[i + 5] == 'e' &&
	  argv[i + 6] == 'r' && argv[i + 7] == 't' && argv[i + 8] == ' ')
	{
	  break;
	}
      else
	fputc (argv[i], stdout);
    }

  while (i++ < size)
    fputc (' ', stdout);
}

/*------------------------------------------------------------------------*/

static int
skip_white_box_test_case (int argc, char ** argv)
{
#ifndef COMPSAT_WHITE
  int i;

  for (i = 1; i < argc; i++)
    if (argv[i][0] == '-' && argv[i][1] == 'W')
      return 1;
#endif
  return 0;
}

/*------------------------------------------------------------------------*/

static int
run (const char *name, int check)
{
  char **argv, *flat_argv, *out, *log;
  int res, i, argc;
  CompSat *compsat;

  parse_opts (name, check, &argc, &argv, &flat_argv);

  printf ("%-12s ", name);
  print_args (53, flat_argv);
  printf (" ...");
  fflush (stdout);

  leaked_bytes = 0;

  if (skip_white_box_test_case (argc, argv))
    {
      res = -1;
    }
  else
    {
      compsat = COMPSAT_new ();
      COMPSAT_set_leak_hook (compsat, &leaked_bytes);

      (void) COMPSAT_main (compsat, argc, argv);

      COMPSAT_delete (compsat);
      out = cons (cons (strdup ("log/"), name), ".out");
      log = cons (cons (strdup ("log/"), name), ".log");
      res = cmp_files (out, log);
      free (out);
      free (log);
    }

  if (leaked_bytes)
    printf (" leaked\n");
  else if (!res)
    printf (" failed\n");
  else if (erase ())
    {
      overwrite ('');
      overwrite (' ');
      overwrite ('');
    }
  else if (res < 0)
    printf (" skipped\n");
  else
    printf (" ok\n");

  fflush (stdout);

  for (i = 0; i < argc; i++)
    free (argv[i]);

  free (argv);
  free (flat_argv);

  return res;
}

/*------------------------------------------------------------------------*/

#define USAGE_COMMON \
"test [<option> ...] [<test-case-name> ...]\n" \
"\n" \
"where <option> is one of the following:\n" \
"\n" \
"  -h | --help      print this command line option summary\n" \
"  -v | --verbose   verbose mode: do not erase output\n" \
"  -f | --fast      run only fast test cases\n"

#define USAGE_TEST_CASE_NAME \
"\n" \
"The <test-case-name> is the prefix of the test cases to be executed.\n" \
"If no prefix is specified, then all test cases are executed.\n"

#ifdef COMPSAT_CHECK
#define USAGE USAGE_COMMON \
"  -c | --check     enable internal consistency checking mode\n" \
USAGE_TEST_CASE_NAME
#else
#define USAGE USAGE_COMMON USAGE_TEST_CASE_NAME
#endif

/*------------------------------------------------------------------------*/

static int
cmp (const void *p, const void *q)
{
  NameWithPriority *a, *b;

  a = (NameWithPriority *) p;
  b = (NameWithPriority *) q;

  if (a->priority < b->priority)
    return -1;

  if (a->priority > b->priority)
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

int
main (int argc, char **argv)
{
  int res, size, count, ok, failed, leaked, matched, skipped, found, check;
  NameWithPriority *testcases, *q;
  char **templates, **p;
  struct dirent *entry;
  char *name;
  DIR *dir;
  int i;

  templates = (char **) malloc (argc * sizeof (char *));
  p = templates;
  check = 0;

  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "-v") || !strcmp (argv[i], "--verbose"))
	{
	  verbosity++;
	}
      else if (!strcmp (argv[i], "-f") || !strcmp (argv[i], "--fast"))
	{
	  fast_test_cases_only = 1;
	}
#ifdef COMPSAT_CHECK
      else if (!strcmp (argv[i], "-c") || !strcmp (argv[i], "--check"))
	{
	  check = 1;
	}
#endif
      else if (!strcmp (argv[i], "-h") || !strcmp (argv[i], "--help"))
	{
	  printf (USAGE);
	  exit (0);
	}
      else if (argv[i][0] == '-')
	{
	  die ("invalid command line option '%s' (try '-h')", argv[i]);
	}
      else
	*p++ = strdup (argv[i]);
    }
  *p = 0;

  printf ("test: updating test case priorities\n");
  res = system ("cd log; make priorities 1>/dev/null");
  if (res)
    die ("error in updating priorities");

  count = 0;
  matched = 0;

  dir = opendir ("log");
  if (!dir)
    die ("could not open 'log' directory");

  testcases = (NameWithPriority *)
    malloc ((size = 1000) * sizeof (NameWithPriority));

  while ((entry = readdir (dir)))
    {
      if (has_opt_prefix (entry->d_name))
	{
	  count++;

	  if (fast_test_cases_only && is_slow (entry->d_name))
	    continue;

	  if (*templates)
	    {
	      found = 0;
	      for (p = templates; !found && *p; p++)
		found = match (entry->d_name, *p);
	    }
	  else
	    found = 1;

	  if (found)
	    {
	      assert (matched < size);
	      name = base (entry->d_name);
	      testcases[matched].name = name;
	      testcases[matched].priority = priority (name);
	      matched++;
	    }
	}
    }
  closedir (dir);
  assert (matched < size);
  testcases[matched].name = 0;
  qsort (testcases, matched, sizeof (NameWithPriority), cmp);

  printf ("test: running %d test cases out of %d\n", matched, count);
  line ();

  ok = failed = skipped = leaked = 0;
  for (q = testcases; q->name; q++)
    {
      res = run (q->name, check);

      if (res > 0)
	ok++;
      else if (res < 0)
	skipped++;
      else
	failed++;

      if (leaked_bytes)
	leaked++;
    }
  if (failed || leaked || (ok && !erase ()))
    line ();
  printf ("%d ok, %d failed, %d leaked\n", ok, failed, leaked);

  for (p = templates; *p; p++)
    free (*p);
  free (templates);

  for (q = testcases; q->name; q++)
    free (q->name);
  free (testcases);

  return 0;
}
